/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifdef HAVE_CONFIG_H
    #include "autoconfig.h"
#endif

#include <boost/pointer_cast.hpp>
#include <errno.h>
#include <string.h>
#include <exception>
#include <iomanip>

#include "common.h"
#include "log.h"
#include "util.h"
#include "exceptions.h"
#include "config_manager.h"
#include "hwdesc.h"
#include "hwstats.h"
#include "storage.h"
#include "database.h"
#include "ds-assistant.h"
#include "database.h"
#include "util.h"
#include "traffic-decode.h"
#include "sqlite3_wrapper.h"

#define INVALID_TASK_ID 0

namespace dSA
{

dS_Assistant* dS_Assistant::m_instance = NULL;

dS_Assistant* dS_Assistant::createInstance()
{
    if (m_instance == NULL)
    {
        m_instance = new dS_Assistant();
        if (m_instance == NULL)
        {
            throw _Exception("could not create dS Assistant class");
        }
    }

    return m_instance;
}

dS_Assistant* dS_Assistant::getInstance()
{
    if (m_instance == NULL)
    {
        throw _Exception("dS Assistant not initialized");
    }

    return m_instance;
}
dS_Assistant::dS_Assistant() : m_shutdown_flag(false),
                               m_running(false),
                               m_task_thread(0),
                               m_long_task_thread(0)
{
    int ret;

    ConfigManager *cfg = ConfigManager::getInstance();
    std::string logdir(cfg->getOption(CFG_LOG_DIR));
    m_logfile_name = logdir +  + "/dsa.log";
    m_logfile = fopen(m_logfile_name.c_str(), "a");
    if (NULL == m_logfile)
    {
        throw _Exception("failed to open logfile \'" + m_logfile_name + "\': " +
                         std::string(strerror(errno)));
    }

    Output2DSID::SetLogdirectory(logdir);
    DSIDLog::ReportingLevel() =
        DSIDLog::FromString((cfg->getOption(CFG_LOG_LEVEL)).c_str());

    Output2FILE::SetStream(m_logfile);
    FILELog::ReportingLevel() =
        FILELog::FromString((cfg->getOption(CFG_LOG_LEVEL)).c_str());
    FILE_LOG(logINFO) << "dS Assistant startup";

    SetNullId(m_own_dsid);

    ret = pthread_mutex_init(&m_task_list_mutex, NULL);
    if (ret != 0)
    {
        throw _Exception("failed to initialize task list mutex: " +
                         std::string(strerror(errno)));
    }

    ret = pthread_mutex_init(&m_long_task_list_mutex, NULL);
    if (ret != 0)
    {
        pthread_mutex_destroy(&m_task_list_mutex);
        throw _Exception("failed to initialize long task list mutex: " +
                         std::string(strerror(errno)));
    }

    ret = pthread_cond_init(&m_wakeup_task_cond, NULL);
    if (ret != 0)
    {
        pthread_mutex_destroy(&m_task_list_mutex);
        pthread_mutex_destroy(&m_long_task_list_mutex);
        throw _Exception("failed to initialize task thread wakeup condition: " +
                         std::string(strerror(errno)));
    }

    ret = pthread_cond_init(&m_wakeup_long_task_cond, NULL);
    if (ret != 0)
    {
        pthread_mutex_destroy(&m_task_list_mutex);
        pthread_mutex_destroy(&m_long_task_list_mutex);
        pthread_cond_destroy(&m_wakeup_task_cond);
        throw _Exception("failed to initialize task thread wakeup condition: " +
                         std::string(strerror(errno)));
    }
}

void dS_Assistant::reopenLog()
{
    Output2FILE::SetStream(NULL);
    if (NULL != m_logfile)
    {
        fclose(m_logfile);
    }
    m_logfile = fopen(m_logfile_name.c_str(), "a");
    if (NULL == m_logfile)
    {
        throw _Exception("failed to re-open logfile \'" + m_logfile_name + "\': " +
                         std::string(strerror(errno)));
    }
    Output2FILE::SetStream(m_logfile);
}

void dS_Assistant::loadDatabaseDefaults()
{
    int apartment_id;
    std::string apartment_id_string;

    // system parameter: apartment id for automatic meter configuration
    try
    {
        ConfigManager::getInstance()->getParameter("digitalstrom.apartmentid",
                apartment_id);
    }
    catch (Exception& ex )
    {
        try
        {
            // migrate old values from database if missing in config file
            apartment_id_string = Database::getInstance()->getConfig("ApartmentId");
            ConfigManager::getInstance()->setParameter("digitalstrom.apartmentid",
                    util::str2int(apartment_id_string));
        }
        catch (Exception& ex )
        {
            // option is neither in config file nor in database
            std::string systemid = util::mac2str();
            char c = systemid.length() ? systemid[systemid.length() - 1] : '0';
            apartment_id = (strtol(&c, NULL, 16) & 7) * 8;

            FILE_LOG(logINFO) << "Set apartment id from system mac address "
                    << systemid << " to " << util::str(apartment_id);
            try
            {
                ConfigManager::getInstance()->setParameter("digitalstrom.apartmentid",
                        apartment_id);
            }
            catch (Exception& ex)
            {
                FILE_LOG(logERROR)
                        << "Failed to store apartment id in configuration: "
                        << ex.getMessage();
            }
        }
    }

    int value;
    bool bvalue;

    // meter configuration: default behavior for checking the device presence
    try
    {
        ConfigManager::getInstance()->getParameter(
                "digitalstrom.meter.pingmode", value);
    }
    catch (Exception& ex)
    {
        try
        {
            value = util::str2int(Database::getInstance()->getConfig(
                    "MeterPingMode"));
        }
        catch (Exception&) {
            value = 3;
        }
        try
        {
            ConfigManager::getInstance()->setParameter(
                    "digitalstrom.meter.pingmode", value);
        }
        catch (Exception& ex)
        {
            FILE_LOG(logERROR)
                << "Failed to store meter.pingmode in configuration: "
                << ex.getMessage();
        }
    }

    // meter configuration: default warning temperature level
    try
    {
        ConfigManager::getInstance()->getParameter(
                "digitalstrom.meter.twarning", value);
    }
    catch (Exception& ex )
    {
        try
        {
            value = util::str2int(Database::getInstance()->getConfig(
                    "MeterTemperatureWarning"));
        }
        catch (Exception&)
        {
            value = 90;
        }
        try
        {
            ConfigManager::getInstance()->setParameter(
                    "digitalstrom.meter.twarning", value);
        }
        catch (Exception& ex)
        {
            FILE_LOG(logERROR)
                << "Failed to store meter.twarning in configuration: "
                << ex.getMessage();
        }
    }

    // meter configuration: default error temperature level
    try
    {
        ConfigManager::getInstance()->getParameter(
                "digitalstrom.meter.tfailure", value);
    }
    catch (Exception& ex ) {
        try
        {
            value = util::str2int(Database::getInstance()->getConfig(
                    "MeterTemperatureError"));
        }
        catch (Exception&) {
            value = 150;
        }
        try
        {
            ConfigManager::getInstance()->setParameter(
                    "digitalstrom.meter.tfailure", value);
        }
        catch (Exception& ex)
        {
            FILE_LOG(logERROR)
                << "Failed to store meter.tfailure in configuration: "
                << ex.getMessage();
        }
    }

    // meter configuration: front led configuration enabled/disabled in idle state
    try
    {
        ConfigManager::getInstance()->getParameter(
                "digitalstrom.meter.led", bvalue);
    }
    catch (Exception& ex ) {
        bvalue = true;
        try
        {
            ConfigManager::getInstance()->setParameter(
                    "digitalstrom.meter.led", bvalue);
        }
        catch (Exception& ex)
        {
            FILE_LOG(logERROR)
                << "Failed to store meter.led in configuration: "
                << ex.getMessage();
        }
    }

    // System parameter: programming mode enabled/disabled
    try
    {
        ConfigManager::getInstance()->getParameter(
                "digitalstrom.programmingmode", bvalue);
    }
    catch (Exception& ex )
    {
        bvalue = true;
        try
        {
            ConfigManager::getInstance()->setParameter(
                    "digitalstrom.programmingmode", bvalue);
        }
        catch (Exception&) {
            FILE_LOG(logERROR) << "Failed to store programmingmode "
                    "parameter in configuration: " << ex.getMessage();
        }
    }

}

void dS_Assistant::connect()
{
    boost::mutex::scoped_lock lock(m_dsa_mutex);

    if (m_dsm_api_wrapper != NULL)
    {
        throw _Exception("connection already exists");
    }

    ConfigManager *cfg = ConfigManager::getInstance();
    m_dsm_api_wrapper = boost::shared_ptr<dSMApiWrapper>(
                    new dSMApiWrapper(cfg->getOption(CFG_DS485_DAEMON_URL)));

    try
    {
        m_dsm_api_wrapper->getOwnDSID(&m_own_dsid);
    }
    catch (Exception ex)
    {
        m_dsm_api_wrapper.reset();
        throw ex;
    }

    m_bus_state = m_dsm_api_wrapper->getBusState();

    m_dsm_api_wrapper->addListener(this);
}

bus_state_t dS_Assistant::getBusState() const
{
    boost::mutex::scoped_lock lock(m_dsa_mutex);
    return m_bus_state;
}

boost::shared_ptr<dSMApiWrapper> dS_Assistant::dS485()
{
    if (m_dsm_api_wrapper == NULL)
    {
        throw _Exception("no active connection to dS485 daemon available");
    }
    return m_dsm_api_wrapper;
}

void dS_Assistant::addTask(boost::shared_ptr<dSTask> task)
{
    if (task == NULL)
    {
        throw _Exception("Will not add invalid NULL task");
    }
    lockTaskList();
    m_task_list.push_front(task);

    pthread_cond_signal(&m_wakeup_task_cond);
    unlockTaskList();
}

long long int dS_Assistant::addLongTask(boost::shared_ptr<dSLongTask> task)
{
    lockLongTaskList();

    long long int task_id = INVALID_TASK_ID;

    try
    {
        if (m_current_task_info != NULL)
        {
            if ((m_current_task_info->getPriority() == Normal) &&
                (task->getPriority() == High))
            {
                if (m_current_task_info->isPausable())
                {
                    FILE_LOG(logDEBUG2) << "pausing current task";
                    m_current_task_info->pause();
                }
                else
                {
                    FILE_LOG(logDEBUG2)
                          << "current task is unpausable and will be cancelled";
                    m_current_task_info->invalidate();
                }
            }
            else if ((m_current_task_info->getPriority() == High) &&
                     (task->getPriority() == High))
            {
                FILE_LOG(logDEBUG2) << "replacing current high prio task";
                m_current_task_info->invalidate();
            }
        }

        task->setTaskStatus(Queued);
        task_id = task->getTaskID();

        if (task_id == INVALID_TASK_ID)
        {
            task_id = Database::getInstance()->saveTask(task);
            task->setTaskID(task_id);
        }
        else
        {
            if (task->getTaskType() !=
                Database::getInstance()->getTaskType(task_id))
            {
                throw _Exception("task type mismatch");
            }
        }

        if (task->getPriority() == High)
        {
            // we agreed to allow only one high prio task for now...
            while (!m_high_prio_task_list.empty())
            {
                boost::shared_ptr<dSLongTask> t = m_high_prio_task_list.front();
                t->invalidate();
                m_high_prio_task_list.pop_front();
            }
            FILE_LOG(logDEBUG2) << "adding high priority task";
            m_high_prio_task_list.push_front(task);
        }
        else
        {
            FILE_LOG(logDEBUG2) << "adding normal priority task";
            m_long_task_list.push_front(task);
        }
        pthread_cond_signal(&m_wakeup_long_task_cond);
    }
    catch (Exception ex)
    {
        unlockLongTaskList();
        throw ex;
    }

    unlockLongTaskList();
    return task_id;
}

void dS_Assistant::scheduleTask(boost::shared_ptr<dSTask> task,
                                unsigned int seconds)
{
    FILE_LOG(logDEBUG2) << "scheduleTask called with delay parameter "
                        << seconds;

    lockTaskList();
    try
    {
        boost::shared_ptr<TaskScheduler> ts(new TaskScheduler(task, seconds));
        FILE_LOG(logDEBUG2) << "scheduleTask: scheduling task in " << seconds
                            << " seconds";
        Timer::getInstance()->addSubscriber(ts);
    }
    catch (Exception ex)
    {
        unlockTaskList();
        throw ex;
    }
    unlockTaskList();
}

void dS_Assistant::scheduleLongTask(boost::shared_ptr<dSLongTask> task,
                           unsigned int seconds)
{
    boost::shared_ptr<LongTaskScheduler> ts;

    FILE_LOG(logDEBUG2) << "scheduleLongTask called with delay parameter "
                        << seconds;

    lockLongTaskList();
    try
    {
        // check and clean up stale tasks
        std::list<boost::weak_ptr<LongTaskScheduler> >::iterator wi =
            m_scheduled_task_list.begin();
        while (wi != m_scheduled_task_list.end())
        {
            boost::weak_ptr<LongTaskScheduler> tw = *wi;
            if (!tw.lock())
            {
                FILE_LOG(logDEBUG2)
                                << "scheduleLongTask: cleaned up stale task";
                wi = m_scheduled_task_list.erase(wi);
                continue;
            }
            wi++;
        }

        task->setTaskStatus(Queued);
        task->setQueuedTime(time(NULL));
        ts = boost::shared_ptr<LongTaskScheduler> (new LongTaskScheduler(task,
            seconds));
        boost::weak_ptr<LongTaskScheduler> tsw(ts);
        m_scheduled_task_list.push_back(tsw);
    }
    catch (Exception ex)
    {
        unlockLongTaskList();
        throw ex;
    }
    unlockLongTaskList();

    FILE_LOG(logDEBUG2) << "scheduleLongTask: scheduling task with id "
                        << task->getTaskID();
    Timer::getInstance()->addSubscriber(ts);
}

///****************************************************************************

void dS_Assistant::busStateCallback(bus_state_t state)
{
    m_bus_state = state;
    FILE_LOG(logINFO) << "dS485 Bus State: " << state;
    try
    {
        /// \todo do not access database in callbacks
        Database::getInstance()->updateDSMPresence(false);
    }
    catch (SQLite3Exception sx)
    {
        FILE_LOG(logERROR) << "Could not update dSM presence: " <<
            sx.getMessage();
        if (SQLite3::isFatal(sx.getErrorCode()))
        {
            FILE_LOG(logERROR) << "Unrecoverable SQLite error, terminating dSA";
            g_shutdown_flag = SHUTDOWN_SQLITE;
        }
    }
}


void dS_Assistant::busChangeCallback(dsid_t *id, int flag)
{
    try
    {
        if (!DsmApiIsdSM(*id))
        {
            return;
        }

        FILE_LOG(logDEBUG1) << "dS485 Bus Change: " << util::str(*id) <<
                               " - " << flag;

        boost::shared_ptr<dSM> dsm;
        try
        {
            dsm = Storage::getInstance()->getDSM(*id);
        }
        catch (SQLite3Exception &sx)
        {
            if (SQLite3::isFatal(sx.getErrorCode()))
            {
                FILE_LOG(logERROR) <<
                    "Unrecoverable SQLite error, terminating dSA";
                g_shutdown_flag = SHUTDOWN_SQLITE;
                return;
            }
            dsm = boost::shared_ptr<dSM>(new dSM(*id));
        }
        catch (Exception &ex)
        {
            dsm = boost::shared_ptr<dSM>(new dSM(*id));
        }

        if (flag)
        {
            DSID_LOG(logINFO, util::str(*id)) << "left";
            dsm->setPresent(0);
        }
        else
        {
            DSID_LOG(logINFO, util::str(*id)) << "joined";
            dsm->setPresent(1);
            boost::shared_ptr<GetDSMInfoTask> task(
                    new GetDSMInfoTask(dsm->getDSID()));
            addTask(boost::dynamic_pointer_cast<dSTask>(task));
        }
        /// \todo do not write to db from callbacks
        Storage::getInstance()->addDSM(dsm);
    }
    catch (SQLite3Exception sx)
    {
        FILE_LOG(logERROR) << "Exception in busChangeCallback: " <<
                              sx.getMessage();
        if (SQLite3::isFatal(sx.getErrorCode()))
        {
            FILE_LOG(logERROR) << "Unrecoverable SQLite error, terminating dSA";
            g_shutdown_flag = SHUTDOWN_SQLITE;
        }
    }
    catch (Exception ex)
    {
        FILE_LOG(logERROR) << "Exception in busChangeCallback: " <<
                              ex.getMessage();
    }
    catch (std::exception e)
    {
        FILE_LOG(logERROR) << "Exception in busChangeCallback: " << e.what();
    }
    catch (...)
    {
        FILE_LOG(logERROR) << "Exception in busChangeCallback";
    }
}

///****************************************************************************

void dS_Assistant::deviceAddedCallback(uint8_t error_code,
                                       dsid_t src_dsm_dsid,
                                       dsid_t dest_dsm_dsid,
                                       uint16_t device_id,
                                       uint16_t zone_id,
                                       uint16_t vendor_id,
                                       uint32_t serial_number)
{
    dsid_t dsid;
    if (DsmApiExpandDeviceDSID(vendor_id, serial_number, &dsid) != 0)
    {
        FILE_LOG(logDEBUG2) << "Ignoring unidentified device on dSM " <<
                               util::str(src_dsm_dsid);
        return;
    }

    try
    {
        boost::shared_ptr<Device> dsd =
                        Storage::getInstance()->getDevice(dsid, src_dsm_dsid);
        dsd->setState(1);
        Storage::getInstance()->addDevice(dsd);
    }
    catch (SQLite3Exception sx)
    {
        FILE_LOG(logERROR) << "Could not load device " << util::str(dsid) <<
            ": " << sx.getMessage();
        if (SQLite3::isFatal(sx.getErrorCode()))
        {
            FILE_LOG(logERROR) << "Unrecoverable SQLite error, terminating dSA";
            g_shutdown_flag = SHUTDOWN_SQLITE;
            return;
        }
    }
    catch (Exception ex)
    {
        // exception is OK if device was not found
    }

    DSID_LOG(logINFO, util::str(dsid)) << "active";
    FILE_LOG(logINFO) << "Added device with dSID " <<
            util::str(dsid) << " and address " << util::str(device_id) <<
            " on dSM " << util::str(src_dsm_dsid);

    boost::shared_ptr<GetDeviceInfoTask> task(
            new GetDeviceInfoTask(src_dsm_dsid, device_id));
    addTask(boost::dynamic_pointer_cast<dSTask>(task));
}

void dS_Assistant::deviceRemovedCallback(uint8_t error_code,
                                         dsid_t src_dsm_dsid,
                                         dsid_t dest_dsm_dsid,
                                         uint16_t device_id,
                                         uint16_t zone_id,
                                         uint16_t vendor_id,
                                         uint32_t serial_number)
{
    dsid_t dsid;
    if (DsmApiExpandDeviceDSID(vendor_id, serial_number, &dsid) != 0)
    {
        FILE_LOG(logDEBUG2) << "Ignoring unidentified device on dSM " <<
                               util::str(src_dsm_dsid);
        return;
    }

    try
    {
        boost::shared_ptr<Device> dsd =
                        Storage::getInstance()->getDevice(dsid, src_dsm_dsid);
        dsd->setState(0);
        Storage::getInstance()->addDevice(dsd);
    }
    catch (SQLite3Exception sx)
    {
        FILE_LOG(logERROR) << "Could not load device " << util::str(dsid) <<
            ": " << sx.getMessage();
        if (SQLite3::isFatal(sx.getErrorCode()))
        {
            FILE_LOG(logERROR) << "Unrecoverable SQLite error, terminating dSA";
            g_shutdown_flag = SHUTDOWN_SQLITE;
            return;
        }
    }
    catch (Exception ex)
    {
        // exception is expected if device was not found
    }
    DSID_LOG(logINFO, util::str(dsid)) << "inactive";
    FILE_LOG(logINFO) << "Removed device with dSID " <<
            util::str(dsid) << " and address " << util::str(device_id) <<
            " on dSM " << util::str(src_dsm_dsid);
}

void dS_Assistant::deviceStatisticsCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t offset, uint16_t value)
{
    try
    {
        boost::shared_ptr<const Device> dsdo =
                Storage::getInstance()->getDevice(device_id, src_dsm_dsid);
        DSID_PREFIX_LOG(logDEBUG, dsdo->getDSID(), "stat_") <<
            "statistic offset:" << util::hex(offset,2) <<
            ", value:" << util::hex(value,4);
    }
    catch (SQLite3Exception sx)
    {
        FILE_LOG(logERROR) << "Could not load device : " << sx.getMessage();
        if (SQLite3::isFatal(sx.getErrorCode()))
        {
            FILE_LOG(logERROR) << "Unrecoverable SQLite error, terminating dSA";
            g_shutdown_flag = SHUTDOWN_SQLITE;
        }
    }
    catch (Exception ex)
    {
        FILE_LOG(logWARNING) << "Device statistics callback: "
                             << ex.getMessage();
    }
}

void dS_Assistant::devicePingCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid,
        uint16_t device_id, uint8_t present)
{
}

void dS_Assistant::deviceActionCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint16_t zone_id, uint8_t group_id, uint8_t button_nr,
        uint8_t click_type)
{
}

void dS_Assistant::deviceLocalActionCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint16_t zone_id, uint8_t state)
{
}

void dS_Assistant::deviceConfigCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t config_class, uint8_t config_index, uint8_t value)
{
    try
    {
        boost::shared_ptr<const Device> dsdo =
                    Storage::getInstance()->getDevice(device_id, src_dsm_dsid);
        DSID_LOG(logINFO, util::str(dsdo->getDSID())) <<
            "device config class:" << util::hex(config_class) <<
            ", index:" << util::hex(config_index, 2) <<
            ", value:" << util::hex(value, 4);
    }
    catch (SQLite3Exception sx)
    {
        FILE_LOG(logERROR) << "Could not load device: " << sx.getMessage();
        if (SQLite3::isFatal(sx.getErrorCode()))
        {
            FILE_LOG(logERROR) << "Unrecoverable SQLite error, terminating dSA";
            g_shutdown_flag = SHUTDOWN_SQLITE;
        }
    }
    catch (Exception ex)
    {
        FILE_LOG(logWARNING) << "Device config callback: "
                             << ex.getMessage();
    }
}

void dS_Assistant::deviceTestTransmissionQualityCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t downstream_quality, uint16_t upstream_quality)
{
    try
    {
        boost::shared_ptr<const Device> dsdo =
            Storage::getInstance()->getDevice(device_id, src_dsm_dsid);
        DSID_LOG(logINFO, util::str(dsdo->getDSID())) <<
            "test quality down:" << util::hex(downstream_quality,2) <<
            ", up:" << util::hex(upstream_quality,2);
    }
    catch (SQLite3Exception sx)
    {
        FILE_LOG(logERROR) << "Could not load device: " << sx.getMessage();
        if (SQLite3::isFatal(sx.getErrorCode()))
        {
            FILE_LOG(logERROR) << "Unrecoverable SQLite error, terminating dSA";
            g_shutdown_flag = SHUTDOWN_SQLITE;
        }
    }
    catch (Exception ex)
    {
        FILE_LOG(logWARNING) << "Device test transmission quality callback: "
                             << ex.getMessage();
    }
}

///****************************************************************************

void dS_Assistant::commUpstreamShortCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t circuit_id, uint8_t is_resend, uint8_t is_sensor,
        uint8_t key_type, uint8_t click_type,
        uint8_t quality, uint8_t flags, uint8_t cflags)
{
}

void dS_Assistant::commUpstreamLongCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t circuit_id, uint8_t mainqualifier, uint8_t subqualifier,
        uint32_t data, uint8_t quality, uint8_t flags, uint8_t cflags)
{
}

void dS_Assistant::commDownstreamShortCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command)
{
}

void dS_Assistant::commDownstreamLong0Callback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command)
{
}

void dS_Assistant::commDownstreamLong8Callback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command,
        uint8_t data)
{
}

void dS_Assistant::commDownstreamLong16Callback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command,
        uint16_t data)
{
}

///****************************************************************************

void dS_Assistant::lockTaskList()
{
    int ret = pthread_mutex_lock(&m_task_list_mutex);
    if (ret != 0)
    {
        throw _Exception("Failed to lock task list mutex: " +
                         std::string(strerror(errno)));
    }
}

void dS_Assistant::unlockTaskList()
{
    int ret = pthread_mutex_unlock(&m_task_list_mutex);
    if (ret != 0)
    {
        throw _Exception("Failed to unlock task list mutex: " +
                         std::string(strerror(errno)));
    }
}

void dS_Assistant::lockLongTaskList() const
{
    int ret = pthread_mutex_lock(&m_long_task_list_mutex);
    if (ret != 0)
    {
        throw _Exception("Failed to lock long task list mutex: " +
                         std::string(strerror(errno)));
    }
}

void dS_Assistant::unlockLongTaskList() const
{
    int ret = pthread_mutex_unlock(&m_long_task_list_mutex);
    if (ret != 0)
    {
        throw _Exception("Failed to unlock long task list mutex: " +
                         std::string(strerror(errno)));
    }
}

void dS_Assistant::run()
{
    int ret;

    if (m_running)
    {
        throw _Exception("Task threads are already running");
    }

    loadDatabaseDefaults();
    connect();

    m_running = true;

    try
    {
        clearPendingTasks();
    }
    catch (Exception ex)
    {
        m_running = false;
        throw ex;
    }

    ret = pthread_create(&m_task_thread, NULL, dS_Assistant::staticThreadProc,
                         this);
    if (ret != 0)
    {
        disconnect();
        m_running = false;
        throw _Exception("Could not start task thread");
    }

    ret = pthread_create(&m_long_task_thread, NULL,
                         dS_Assistant::staticLongTaskThreadProc, this);
    if (ret != 0)
    {
        pthread_join(m_task_thread, NULL);
        disconnect();
        m_running = false;
        throw _Exception("Could not start long task thread");
    }

    try
    {
        m_comm_monitor_task = boost::shared_ptr<BasicMonitorTask>
            (new BasicMonitorTask());
        addTask(boost::dynamic_pointer_cast<dSTask>
            (m_comm_monitor_task));

        int max_db_size;
        try
        {
            ConfigManager::getInstance()->getParameter(
                    "database.maxfilesize", max_db_size);
        }
        catch (Exception& ex) {
            max_db_size = DATABASE_MAX_FILESIZE;
        }

        m_db_cleanup_task = boost::shared_ptr<DatabaseCleanupTask>
            (new DatabaseCleanupTask(max_db_size));
        scheduleTask(boost::dynamic_pointer_cast<dSTask>
            (m_db_cleanup_task), 300);
    }
    catch (Exception ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }

    try
    {
        Database::getInstance()->updateTaskStatus(Running, Cancelled);
        Database::getInstance()->updateDSMPresence(false);
#ifdef ADVANCED_TASK_MANAGEMENT
        scan(false); // synchronous scan

         boost::shared_ptr<std::list<boost::shared_ptr<dSLongTask> > > tasks =
            Database::getInstance()->getTasksByStatus(Queued);

        if (tasks->empty())
        {
            return;
        }

        lockLongTaskList();
        while (!tasks->empty())
        {
            boost::shared_ptr<dSLongTask> task = tasks->front();
            tasks->pop_front();
            m_long_task_list.push_front(task);

        }
        pthread_cond_signal(&m_wakeup_long_task_cond);
        unlockLongTaskList();
#else
        Database::getInstance()->removeTasksByStatus(Queued);
        scan(true); // asynchronous scan
#endif
    }
    catch (SQLite3Exception sx)
    {
        if (SQLite3::isFatal(sx.getErrorCode())) {
            throw sx;
        }
        FILE_LOG(logERROR) << sx.getMessage();
    }
    catch (Exception ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
}

void dS_Assistant::taskThread()
{
    boost::shared_ptr<dSTask> task;
    bool terminate = false;

    lockTaskList();
    while (!m_shutdown_flag && !terminate)
    {
        if (m_task_list.empty())
        {
            pthread_cond_wait(&m_wakeup_task_cond, &m_task_list_mutex);
            continue;
        }

        task = m_task_list.back();
        m_task_list.pop_back();

        unlockTaskList();

        try
        {
            if ((task != NULL) && (task->isValid()))
            {
                task->run(this);
            }
        }
        catch (SQLite3Exception sx)
        {
            FILE_LOG(logERROR) << "Task thread caught exception: " <<
                sx.getMessage();
            sx.printStackTrace();
            if (SQLite3::isFatal(sx.getErrorCode()))
            {
                FILE_LOG(logERROR) << "Unrecoverable SQLite error, terminating";
                terminate = true;
            }
        }
        catch (Exception ex)
        {
            FILE_LOG(logERROR) << "Task thread caught exception: " <<
                                  ex.getMessage();
            ex.printStackTrace();
        }
        catch (std::exception e)
        {
            FILE_LOG(logERROR) << "Task thread caught std exception: " <<
                                  e.what();
        }
        catch (...)
        {
            FILE_LOG(logERROR) << "Task thread caught unknown exception";
        }

        task.reset();

        if (!m_shutdown_flag && !terminate)
        {
            lockTaskList();
        }
    }
    unlockTaskList();
    if (terminate)
    {
        FILE_LOG(logERROR) << "Unrecoverable SQLite error, terminating dSA";
        g_shutdown_flag = SHUTDOWN_SQLITE;
    }
}

void *dS_Assistant::staticThreadProc(void *arg)
{
    dS_Assistant *instance = (dS_Assistant *)arg;
    instance->taskThread();
    pthread_exit(NULL);
    return NULL;
}

void dS_Assistant::longTaskThread()
{
    boost::shared_ptr<dSLongTask> task;
    bool terminate = false;

    lockLongTaskList();
    while (!m_shutdown_flag && !terminate)
    {
        if ((m_paused_task_list.empty()) &&
            (m_long_task_list.empty()) &&
            (m_high_prio_task_list.empty()))
        {
            FILE_LOG(logDEBUG2) << "no long tasks pending, sleeping...";
            pthread_cond_wait(&m_wakeup_long_task_cond,
                              &m_long_task_list_mutex);
            continue;
        }

        // high prio tasks will always be sorted in front by the add function
        if (!m_high_prio_task_list.empty())
        {
            FILE_LOG(logDEBUG2) << "queueing high priority task";
            task = m_high_prio_task_list.back();
            m_high_prio_task_list.pop_back();
        }

        // if no high priority task is queued, proceed with the paused tasks
        if ((task == NULL) && (!m_paused_task_list.empty()))
        {
            FILE_LOG(logDEBUG2) << "queueing paused task";
            task = m_paused_task_list.back();
            m_paused_task_list.pop_back();
        }

        if ((task == NULL) && (!m_long_task_list.empty()))
        {
            FILE_LOG(logDEBUG2) << "queueing regular task";
            task = m_long_task_list.back();
            m_long_task_list.pop_back();
        }

        m_current_task_info = boost::dynamic_pointer_cast<dSLongTaskInfo>(task);
        unlockLongTaskList();

        try
        {
            if (task->isValid())
            {
                Database::getInstance()->updateTaskStatus(task->getTaskID(),
                                                          Running);
                task->run(this);
                Database::getInstance()->updateTaskStatus(task->getTaskID(),
                                                        task->getTaskStatus());
                if (task->isValid() && task->isPaused())
                {
                    m_paused_task_list.push_front(task);
                }
            }
        }
        catch (SQLite3Exception sx)
        {
            FILE_LOG(logERROR) << "Long task thread caught exception: " <<
                sx.getMessage();
            sx.printStackTrace();
            if (SQLite3::isFatal(sx.getErrorCode()))
            {
                terminate = true;
            }
        }
        catch (Exception ex)
        {
            FILE_LOG(logERROR) << "Long task thread caught exception: " <<
                ex.getMessage();
            ex.printStackTrace();
        }
        catch (std::exception e)
        {
            FILE_LOG(logERROR) << "Long task thread caught exception: " <<
                e.what();
        }
        catch (...)
        {
            FILE_LOG(logERROR) << "Long task thread caught unknown exception";
        }

        task.reset();

        if (!m_shutdown_flag && !terminate)
        {
            lockLongTaskList();
            m_current_task_info.reset();
        }
    }
    unlockLongTaskList();

    if (terminate)
    {
        FILE_LOG(logERROR) << "Unrecoverable SQLite error, terminating dSA";
        g_shutdown_flag = SHUTDOWN_SQLITE;
    }
}

void *dS_Assistant::staticLongTaskThreadProc(void *arg)
{
    dS_Assistant *instance = (dS_Assistant *)arg;
    instance->longTaskThread();
    pthread_exit(NULL);
    return NULL;
}

void dS_Assistant::scan(bool async)
{
    Database::getInstance()->updateDSMPresence(false);
    boost::shared_ptr<GetBusMembersTask> task(new GetBusMembersTask(async));
    if (async)
    {
        addTask(boost::dynamic_pointer_cast<dSTask>(task));
    }
    else
    {
        task->run(this);
    }

    boost::shared_ptr<CheckDSMThread> dsmtask(new CheckDSMThread(1800));
    scheduleLongTask(boost::dynamic_pointer_cast<dSLongTask>(dsmtask), 60);
}

void dS_Assistant::clearPendingTasks()
{
    lockTaskList();
    while (!m_task_list.empty())
    {
        m_task_list.pop_front();
    }
    unlockTaskList();

    lockLongTaskList();
    while (!m_high_prio_task_list.empty())
    {
        m_high_prio_task_list.pop_front();
    }

    while (!m_long_task_list.empty())
    {
        m_long_task_list.pop_front();
    }

    while (!m_paused_task_list.empty())
    {
        m_paused_task_list.pop_front();
    }

    try
    {
        while (!m_scheduled_task_list.empty())
        {
            boost::weak_ptr<LongTaskScheduler> tw = m_scheduled_task_list.front();
            if (boost::shared_ptr<LongTaskScheduler> t = tw.lock())
            {
                Timer::getInstance()->removeSubscriber(t);
            }
            m_scheduled_task_list.pop_front();
        }
    }
    catch (Exception ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
    unlockLongTaskList();
}

boost::shared_ptr<const dSLongTaskInfo> dS_Assistant::getCurrentTaskInfo() const
{
    lockLongTaskList();
    boost::shared_ptr<const dSLongTaskInfo> ret = m_current_task_info;
    unlockLongTaskList();
    return ret;
}

boost::shared_ptr<std::list<boost::shared_ptr<const dSLongTaskInfo> > >
                                        dS_Assistant::getTaskList() const
{
    lockLongTaskList();
    boost::shared_ptr<std::list<boost::shared_ptr<const dSLongTaskInfo> > >
        taskList(new std::list<boost::shared_ptr<const dSLongTaskInfo> >);

    if (m_current_task_info != NULL)
    {
        taskList->push_back(m_current_task_info);
    }

    for (std::list<boost::shared_ptr<dSLongTask> >::const_iterator
            it = m_high_prio_task_list.begin();
            it != m_high_prio_task_list.end(); it++)
    {
        boost::shared_ptr<const dSLongTaskInfo> task =
                            boost::dynamic_pointer_cast<dSLongTaskInfo>(*it);
        taskList->push_back(task);
    }

    for (std::list<boost::shared_ptr<dSLongTask> >::const_iterator
            it = m_paused_task_list.begin();
            it != m_paused_task_list.end(); it++)
    {
        boost::shared_ptr<const dSLongTaskInfo> task =
                            boost::dynamic_pointer_cast<dSLongTaskInfo>(*it);
        taskList->push_back(task);
    }

    for (std::list<boost::shared_ptr<dSLongTask> >::const_iterator
            it = m_long_task_list.begin(); it != m_long_task_list.end(); it++)
    {
        boost::shared_ptr<const dSLongTaskInfo> task =
                            boost::dynamic_pointer_cast<dSLongTaskInfo>(*it);
        taskList->push_back(task);
    }

    for (std::list<boost::weak_ptr<LongTaskScheduler> >::const_iterator
            it = m_scheduled_task_list.begin();
            it != m_scheduled_task_list.end(); it++)
    {
        boost::shared_ptr<LongTaskScheduler> s = (*it).lock();
        if (s)
        {
            boost::shared_ptr<const dSLongTaskInfo> task = s->getTask();
            taskList->push_back(task);
        }
    }

    unlockLongTaskList();

    return taskList;
}

void dS_Assistant::pauseTask(long long int task_id)
{
    bool found = false;
    lockLongTaskList();

    if ((m_current_task_info != NULL) &&
        ((m_current_task_info->getTaskID() == task_id) ||
         (m_current_task_info->getParentTaskID() == task_id)))
    {
        found = true;
        if (m_current_task_info->isPausable())
        {
            m_current_task_info->pause();
        }
        else
        {
            unlockLongTaskList();
            throw _Exception("task does not support pausing");
        }
    }

    try
    {
        if (found)
        {
            Database::getInstance()->updateTaskStatus(task_id, Cancelled);
        }
    }
    catch (Exception ex)
    {
        FILE_LOG(logDEBUG) << ex.getMessage();
        unlockLongTaskList();
        throw ex;
    }

    unlockLongTaskList();

    if (!found)
    {
        throw _Exception("task with id " + util::str(task_id) + " not found");
    }
}

void dS_Assistant::cancelTask(long long int task_id)
{
    bool found = false;

    std::list<long long int> parents;

    lockLongTaskList();
    try
    {
        std::list<boost::shared_ptr<dSLongTask> >::iterator it =
            m_long_task_list.begin();

        while (it != m_long_task_list.end())
        {
            boost::shared_ptr<dSLongTask> task = *it;
            if ((task->getTaskID() == task_id) ||
                (task->getParentTaskID() == task_id))
            {
                found = true;
                task->invalidate();
                it = m_long_task_list.erase(it);
                if (task->getParentTaskID() != INVALID_TASK_ID)
                {
                    parents.push_front(task->getParentTaskID());
                }
                continue;
            }
            it++;
        }

        it = m_high_prio_task_list.begin();
        while (it != m_high_prio_task_list.end())
        {
            boost::shared_ptr<dSLongTask> task = *it;
            if ((task->getTaskID() == task_id) ||
                (task->getParentTaskID() == task_id))
            {
                found = true;
                task->invalidate();
                it = m_high_prio_task_list.erase(it);
                if (task->getParentTaskID() != INVALID_TASK_ID)
                {
                    parents.push_front(task->getParentTaskID());
                }
                continue;
            }
            it++;
        }

        it = m_paused_task_list.begin();
        while (it != m_paused_task_list.end())
        {
            boost::shared_ptr<dSLongTask> task = *it;
            if ((task->getTaskID() == task_id) ||
                (task->getParentTaskID() == task_id))
            {
                found = true;
                task->invalidate();
                it = m_paused_task_list.erase(it);
                if (task->getParentTaskID() != INVALID_TASK_ID)
                {
                    parents.push_front(task->getParentTaskID());
                }
                continue;
            }
            it++;
        }

        std::list<boost::weak_ptr<LongTaskScheduler> >::iterator wi =
            m_scheduled_task_list.begin();
        while (wi != m_scheduled_task_list.end())
        {
            boost::weak_ptr<LongTaskScheduler> tw = *wi;
            if (boost::shared_ptr<LongTaskScheduler> ts = tw.lock())
            {
                if ((ts->getTask()->getTaskID() == task_id) ||
                    (ts->getTask()->getParentTaskID() == task_id))
                {
                    found = true;
                    ts->getTask()->invalidate();
                    Timer::getInstance()->removeSubscriber(ts);
                    wi = m_scheduled_task_list.erase(wi);
                    if (ts->getTask()->getParentTaskID() != INVALID_TASK_ID)
                    {
                        parents.push_front(ts->getTask()->getParentTaskID());
                    }
                    continue;
                }
            }
            else
            {
                wi = m_scheduled_task_list.erase(wi);
                continue;
            }
            wi++;
        }

        if ((m_current_task_info != NULL) &&
            ((m_current_task_info->getTaskID() == task_id) ||
             (m_current_task_info->getParentTaskID() == task_id)))
        {
            found = true;
            if (m_current_task_info->getParentTaskID() != INVALID_TASK_ID)
            {
                parents.push_front(m_current_task_info->getParentTaskID());
            }
            m_current_task_info->invalidate();
        }

        if (found)
        {
            Database::getInstance()->updateTaskStatus(task_id, Cancelled);
        }
    }
    catch (Exception ex)
    {
        FILE_LOG(logDEBUG) << ex.getMessage();
        unlockLongTaskList();
        throw ex;
    }
    catch (std::runtime_error err)
    {
        FILE_LOG(logERROR) << err.what();
    }
    unlockLongTaskList();

    if (!found)
    {
        throw _Exception("task with id " + util::str(task_id) + " not found");
    }

    parents.unique();
    while (!parents.empty())
    {
        try
        {
            this->cancelTask(parents.front());
        }
        catch (Exception ex)
        {
            // parent tasks may not be around anymore
            FILE_LOG(logDEBUG2) << " OK if task now found: " << ex.getMessage();
        }
        parents.pop_front();
    }
}

void dS_Assistant::disconnect()
{
    boost::mutex::scoped_lock lock(m_dsa_mutex);
    if (m_dsm_api_wrapper)
    {
        try
        {
            clearPendingTasks();
        }
        catch (Exception ex)
        {
            FILE_LOG(logERROR) << "Failed to clear pending tasks: " <<
                ex.getMessage();
        }
        m_dsm_api_wrapper->removeListener(this);
        m_dsm_api_wrapper.reset();
    }
}

void dS_Assistant::shutdown()
{
    boost::mutex::scoped_lock lock(m_dsa_mutex);
    m_shutdown_flag = true;

    clearPendingTasks();
    lockTaskList();
    pthread_cond_signal(&m_wakeup_task_cond);
    unlockTaskList();
    if (m_task_thread)
    {
        pthread_join(m_task_thread, NULL);
        m_task_thread = 0;
    }

    lockLongTaskList();
    if (m_current_task_info != NULL)
    {
        m_current_task_info->invalidate();
    }
    pthread_cond_signal(&m_wakeup_long_task_cond);
    unlockLongTaskList();
    if (m_long_task_thread)
    {
        pthread_join(m_long_task_thread, NULL);
        m_long_task_thread = 0;
    }

    lock.unlock();
    disconnect();
}

const dsid_t dS_Assistant::getDSID() const
{
    return m_own_dsid;
}

long long int dS_Assistant::getInvalidTaskID() const
{
    return INVALID_TASK_ID;
}
dS_Assistant::~dS_Assistant()
{
    boost::mutex::scoped_lock lock(m_dsa_mutex);
    m_instance = NULL;
    lock.unlock();
    try
    {
        shutdown();
        disconnect();
        Database::getInstance()->updateTaskStatus(Running, Cancelled);
        Database::getInstance()->updateDSMPresence(false);
#ifndef ADVANCED_TASK_MANAGEMENT
        Database::getInstance()->removeTasksByStatus(Queued);
#endif
    }
    catch (...)
    {
        FILE_LOG(logERROR) << "Exception caught while shutting down, ignoring";
    }

    pthread_mutex_destroy(&m_task_list_mutex);
    pthread_mutex_destroy(&m_long_task_list_mutex);
    pthread_cond_destroy(&m_wakeup_task_cond);
    pthread_cond_destroy(&m_wakeup_long_task_cond);
}

dSTaskInfo::dSTaskInfo() : m_name("Generic task"), m_description(""),
                           m_valid(false) {}

void dSTaskInfo::setDescription(std::string description)
{
    m_description = description;
}

const std::string dSTaskInfo::getName() const
{
    return m_name;
}

const std::string dSTaskInfo::getDescription() const
{
    return m_description;
}

bool dSTaskInfo::isValid() const
{
    return m_valid;
}

void dSTaskInfo::invalidate()
{
    m_valid = false;
}

dSTaskInfo::~dSTaskInfo() {}

dSTask::dSTask() : dSTaskInfo() {}

dSLongTaskInfo::dSLongTaskInfo() : dSTaskInfo(), m_task_id(INVALID_TASK_ID),
                                    m_parent_task_id(INVALID_TASK_ID),
                                    m_progress(0),
                                    m_status_info("unbekannt"),
                                    m_task_status(UnknownStatus),
                                    m_started(time(NULL)),
                                    m_finished(0),
                                    m_queued(0),
                                    m_paused(false),
                                    m_pausable(false),
                                    m_task_priority(Normal) {}

void dSLongTaskInfo::setTaskID(long long int task_id)
{
    m_task_id = task_id;
}

long long int dSLongTaskInfo::getTaskID() const
{
    return m_task_id;
}

void dSLongTaskInfo::setParentTaskID(long long int task_id)
{
    m_parent_task_id = task_id;
}

long long int dSLongTaskInfo::getParentTaskID() const
{
    return m_parent_task_id;
}

unsigned char dSLongTaskInfo::getProgress() const
{
    boost::mutex::scoped_lock lock(m_status_mutex);
    return m_progress;
}

std::string dSLongTaskInfo::getStatusInfo() const
{
    boost::mutex::scoped_lock lock(m_status_mutex);
    return m_status_info;
}

std::string dSLongTaskInfo::getParameters() const
{
    return m_parameters;
}

void dSLongTaskInfo::setTaskStatus(task_status status)
{
    boost::mutex::scoped_lock lock(m_status_mutex);
    m_task_status = status;
    switch (status)
    {
        case Running:
            m_status_info = "läuft gerade";
            break;
        case Queued:
            m_status_info = "gequeued";
            break;
        case Cancelled:
            m_status_info = "abgebrochen";
            break;
        case Finished:
            m_status_info = "beendet";
            break;
        case UnknownStatus:
        default:
            m_status_info = "unbekannt";
            break;
    }
}

task_status dSLongTaskInfo::getTaskStatus() const
{
    boost::mutex::scoped_lock lock(m_status_mutex);
    return m_task_status;
}

std::string dSLongTaskInfo::getTaskStatusText() const
{
    boost::mutex::scoped_lock lock(m_status_mutex);
    if (m_task_status == Running)
    {
        return "running";
    }
    else if (m_task_status == Cancelled)
    {
        return "cancelled";
    }
    else if (m_task_status == Finished)
    {
        return "finished";
    }
    else if (m_task_status == Queued)
    {
        return "queued";
    }
    else
    {
        return "unknown";
    }
}

time_t dSLongTaskInfo::getStartedTime() const
{
    return m_started;
}

void dSLongTaskInfo::setStartedTime(time_t t)
{
    m_started = t;
}

time_t dSLongTaskInfo::getFinishedTime() const
{
    return m_finished;
}

void dSLongTaskInfo::setFinishedTime(time_t t)
{
    m_finished = t;
}

time_t dSLongTaskInfo::getQueuedTime() const
{
    return m_queued;
}

void dSLongTaskInfo::setQueuedTime(time_t t)
{
    m_queued = t;
}

bool dSLongTaskInfo::isPaused() const
{
    return m_paused;
}

void dSLongTaskInfo::pause()
{
    m_paused = true;
}

bool dSLongTaskInfo::isPausable() const
{
    return m_pausable;
}

void dSLongTaskInfo::setPriority(task_priority priority)
{
    m_task_priority = priority;
}

task_priority dSLongTaskInfo::getPriority() const
{
    return m_task_priority;
}

dSLongTask::dSLongTask() : dSLongTaskInfo() {}

TaskScheduler::TaskScheduler(boost::shared_ptr<dSTask> task,
                             unsigned int seconds) :
                             TimerSubscriber(seconds, true),
                             m_task(task)
{}

boost::shared_ptr<dSTask> TaskScheduler::getTask()
{
    return m_task;
}

void TaskScheduler::notify()
{
    dS_Assistant::getInstance()->addTask(m_task);
}



LongTaskScheduler::LongTaskScheduler(boost::shared_ptr<dSLongTask> task,
                                     unsigned int seconds) :
                                     TimerSubscriber(seconds, true),
                                     m_task(task)
{}

boost::shared_ptr<dSLongTask> LongTaskScheduler::getTask()
{
    return m_task;
}

void LongTaskScheduler::notify()
{
    dS_Assistant::getInstance()->addLongTask(m_task);
}

GetBusMembersTask::GetBusMembersTask(bool async) : dSTask(), m_async(async)
{
    m_name = "Get bus members";
    m_description = "Scan the dS485 bus in order to retrieve information "
                    "about all available devices";
    m_valid = true;
}


void GetBusMembersTask::run(dS_Assistant *dsa)
{
    dsid_t device_list[DSM_API_MAX_DEVICES];

    if (!(dsa && m_valid))
    {
        return;
    }

    int devices = dsa->dS485()->getBusMembers(device_list, DSM_API_MAX_DEVICES);
    for (int i = 0; i < devices; ++i)
    {
        dsid_t own_dsid = dsa->getDSID();
        if (IsEqualId(device_list[i], own_dsid))
        {
            continue;
        }

        if (DsmApiIsdSM(device_list[i]))
        {
            boost::shared_ptr<GetDSMInfoTask> task(
                                  new GetDSMInfoTask(device_list[i], m_async));
            if (m_async)
            {
                dsa->addTask(boost::dynamic_pointer_cast<dSTask>(task));
            }
            else
            {
                task->run(dsa);
            }
        }
        else
        {
             FILE_LOG(logWARNING) << "Ignoring unkwnown device with id "
                                  << util::str(device_list[i]);
        }
    }
}

GetDSMInfoTask::GetDSMInfoTask(dsid_t dsid, bool async) : dSTask(),
                                                          m_dsid(dsid),
                                                          m_async(async)
{
    m_name = "Request dSM Information";
    m_description = "Retrieve information for dSM " + util::str(dsid);
    m_valid = true;
}

void GetDSMInfoTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    uint32_t hwv;
    uint32_t armv;
    uint32_t dspv;
    uint16_t apiv;
    dsid_t dsid;
    uint8_t name[21];
    uint32_t cHash, cVersion;
    uint64_t configId;

    dsa->dS485()->dSMInfo(m_dsid, &hwv, &armv, &dspv, &apiv, &dsid, name);
    dsa->dS485()->getConfigurationId(m_dsid, &cHash, &cVersion);
    configId = (uint64_t) cHash << 32 | cVersion;

    DSID_LOG(logINFO, m_dsid) << "[Task DSMInfo] ARM:" << util::hex(armv) <<
            ", DSP:" << util::hex(dspv) <<
            ", HW:" << util::hex(hwv) <<
            ", CFG:" << configId;

    boost::shared_ptr<dSM> dsm;
    try
    {
        dsm = Storage::getInstance()->getDSM(m_dsid);
    }
    catch (SQLite3Exception &sx)
    {
        FILE_LOG(logERROR) << "Could not load dSM: " << sx.getMessage();
        if (SQLite3::isFatal(sx.getErrorCode()))
        {
            throw sx;
        }
        dsm = boost::shared_ptr<dSM>(new dSM(m_dsid));
    }
    catch (Exception &ex)
    {
        dsm = boost::shared_ptr<dSM>(new dSM(m_dsid));
    }

    dsm->setPresent(1);
    dsm->setHWVersion(hwv);
    dsm->setARMSwVersion(armv);
    dsm->setDSPSwVersion(dspv);
    dsm->setAPIVersion(apiv);
    dsm->setName(std::string((const char*)name));
    Storage::getInstance()->addDSM(dsm);

    // Synchronously verify meter configuration
    boost::shared_ptr<StartupDSMTask> task(new StartupDSMTask(m_dsid, false));
    task->run(dsa);

    // The startup task may have restarted the dSM, no use to proceed here now.
    // After the dSM is back and up again we will come back to here anyway.
    dsm = Storage::getInstance()->getDSM(m_dsid);
    if (! dsm->getPresent())
    {
        return;
    }

    if (configId != dsm->getConfigurationId())
    {
        dsm->setConfigurationId(configId);
        Storage::getInstance()->addDSM(dsm);

        boost::shared_ptr<ScanDSMTask> task(new ScanDSMTask(m_dsid, m_async));
        if (m_async)
        {
            dsa->addTask(boost::dynamic_pointer_cast<dSTask>(task));
        }
        else
        {
            task->run(dsa);
        }
    }
}

ScanDSMTask::ScanDSMTask(dsid_t dsid, bool async) : dSTask(), m_dsid(dsid),
                                                    m_async(async)
{
    m_name = "Scanning dSM for devices";
    m_description = "Retrieve information about devices attached to the "
                    "dS-Meter";
    m_valid = true;
}

void ScanDSMTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    boost::shared_ptr<dSM> dsm;
    try
    {
        dsm = Storage::getInstance()->getDSM(m_dsid);
    }
    catch (SQLite3Exception &sx)
    {
        FILE_LOG(logERROR) << "Could not load dSM " << util::str(m_dsid) <<
            ": " << sx.getMessage();
        if (SQLite3::isFatal(sx.getErrorCode()))
        {
            throw sx;
        }
        return;
    }
    catch (Exception &ex)
    {
        FILE_LOG(logERROR) << "[Task ScanDSM] dSM with ID " <<
                util::str(m_dsid) << " not found in database";
        return;
    }

    uint16_t num;
    dsa->dS485()->getDevicesCountInZone(m_dsid, 0, &num);
    dsm->setDeviceCount(num);
    dsa->dS485()->getActiveDevicesCountInZone(m_dsid, 0, &num);
    dsm->setActiveDeviceCount(num);
    dsa->dS485()->getInactiveDevicesCountInZone(m_dsid, 0, &num);
    dsm->setInactiveDeviceCount(num);

    uint8_t count;
    dsa->dS485()->getZoneCount(m_dsid, &count);

    for (uint8_t i = 0; i < count; i++)
    {
        uint16_t zone_id;
        dsa->dS485()->getZoneInfo(m_dsid, i, &zone_id, NULL, NULL, NULL);
        dsm->setZoneId(i, zone_id);
    }

    // update storage
    Storage::getInstance()->addDSM(dsm);

    // schedule device scanner task
    boost::shared_ptr<GetAllDevicesTask> task(new GetAllDevicesTask(m_dsid));
    if (m_async)
    {
        dsa->addTask(boost::dynamic_pointer_cast<dSTask>(task));
    }
    else
    {
        task->run(dsa);
    }
}
StartupDSMTask::StartupDSMTask(dsid_t dsid, bool async) : dSTask(),
                                                    m_dsid(dsid), m_async(async)
{
    m_name = "dSM startup handler";
    m_description = "Check dSM configuration on startup";
    m_valid = true;
}

void StartupDSMTask::run(dS_Assistant *dsa)
{
    bool restartRequired = false;
    boost::shared_ptr<dSM> dsmconfig;

    try
    {
        dsmconfig = Storage::getInstance()->getDSM(m_dsid);
    }
    catch (SQLite3Exception &sx)
    {
        FILE_LOG(logERROR) << "Could not load dSM " << util::str(m_dsid) <<
            ": " << sx.getMessage();
        if (SQLite3::isFatal(sx.getErrorCode()))
        {
            throw sx;
        }
        return;
    }
    catch (Exception &ex)
    {
        FILE_LOG(logERROR) << "[Task StartupDSM] dSM with ID " <<
                util::str(m_dsid) << " not found in database";
        return;
    }

    if (!(dsa && m_valid))
    {
        return;
    }

    for (int retry = 0; retry < 3; retry ++ )
    {

        // 1. Verify apartment id
        try
        {
            uint8_t aid;
            int dbid;
            ConfigManager::getInstance()->getParameter(
                    "digitalstrom.apartmentid", dbid);
            dsa->dS485()->getApartmentId(m_dsid, &aid);
            if (aid != dbid)
            {
                dsa->dS485()->setApartmentId(m_dsid, dbid);
                restartRequired = true;
                DSID_LOG(logINFO, m_dsid) << "[Task StartupDSM] ApartmentId "
                "reconfiguration from " << util::str(aid) <<
                " to " << util::str(dbid);
            }
            dsmconfig->setApartmentId(aid);
        }
        catch(Exception ex)
        {
            DSID_LOG(logERROR, m_dsid) << "[Task StartupDSM] "
                    "Reconfigure ApartmentId: " << ex.getMessage();
            continue;
        }

        // 2. Verify meter settings, get default modes from db
        try
        {
            uint8_t flags, newflags;
            int pmode;
            bool led_on;
            bool progmode;
            ConfigManager::getInstance()->getParameter(
                    "digitalstrom.meter.led", led_on);
            ConfigManager::getInstance()->getParameter(
                    "digitalstrom.meter.pingmode", pmode);
            ConfigManager::getInstance()->getParameter(
                    "digitalstrom.programmingmode", progmode);
            dsa->dS485()->getPropertyFlags(m_dsid, &flags);
            newflags = flags;

            uint8_t fvalue = led_on ? 1 : 0;
            if ((flags & 1) != fvalue)
            {
                newflags &= ~1;
                newflags |= fvalue;
                restartRequired = true;
            }
            if (((flags >> 1) & 3) != pmode)
            {
                newflags &= ~(3 << 1);
                newflags |= (pmode << 1);
                restartRequired = true;
                DSID_LOG(logDEBUG, m_dsid) << "[Task StartupDSM] Ping Mode "
                        "reconfiguration " << util::str((flags >> 1) & 3) <<
                        " => " << util::str(pmode);
            }
            fvalue = progmode ? 0 : 8;
            if ((flags & 8) != fvalue)
            {
                newflags &= ~8;
                newflags |= fvalue;
            }

            if (newflags != flags)
            {
                dsa->dS485()->setPropertyFlags(m_dsid, newflags);
                DSID_LOG(logINFO, m_dsid) << "[Task StartupDSM] "
                        "Reconfigure dSM Flags " <<
                        util::hex(flags, 2) << " => " << util::hex(newflags, 2);
            }
        }
        catch(Exception ex)
        {
            DSID_LOG(logERROR, m_dsid) << "[Task StartupDSM] "
                    "Reconfigure dSM Flags: " << ex.getMessage();
            continue;
        }

        // 3. Verify temperature limits, get defaults from db (90/150 instead of 70/90)
        try
        {
            uint8_t wt, et;
            int warnlevel;
            int errorlevel;
            ConfigManager::getInstance()->getParameter("digitalstrom.meter.twarning", warnlevel);
            ConfigManager::getInstance()->getParameter("digitalstrom.meter.tfailure", errorlevel);
            dsa->dS485()->getPropertyTemperatureLimit(m_dsid, &wt, &et);
            if (wt != warnlevel || et != errorlevel)
            {
                dsa->dS485()->setPropertyTemperatureLimit(m_dsid, warnlevel, errorlevel);
                restartRequired = true;
                DSID_LOG(logINFO, m_dsid) << "[Task StartupDSM] Temperature level "
                "reconfiguration from " << util::str(wt) << "/" << util::str(et) <<
                " to " << util::str(warnlevel) << "/" << util::str(errorlevel);
            }
        }
        catch(Exception ex)
        {
            DSID_LOG(logERROR, m_dsid) << "[Task StartupDSM] "
                    "Reconfigure Temperature Limits: " << ex.getMessage();
            continue;
        }

        // 4. Check: Communication Log
        dsa->dS485()->setCommunicationLogging(m_dsid, 0xff);

        // Restart dSM if required, reset present flags
        try
        {
            if (restartRequired)
            {
                dsa->dS485()->restartMeter(m_dsid);
                dsmconfig->setPresent(0);
            }
        }
        catch(Exception ex)
        {
            DSID_LOG(logERROR, m_dsid) << "[Task StartupDSM] "
                    "Restart dSM failed: " << ex.getMessage();
            continue;
        }

        break;
    }

    Storage::getInstance()->addDSM(dsmconfig);
}

SetDSMApartmentIdTask::SetDSMApartmentIdTask(dsid_t dsid, int apartment_id,
        bool async) :
        dSTask(), m_dsid(dsid), m_apartment_id(apartment_id), m_async(async)
{
    m_name = "Set apartment id handler";
    m_description = "Set dSM apartment id and restart";
    m_valid = true;
}

void SetDSMApartmentIdTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    boost::shared_ptr<std::list<dsid_t> > dsms;
    boost::shared_ptr<std::list<dsid_t> > dsm_list(new std::list<dsid_t>);
    if (IsNullId(m_dsid))
    {
        dsms = Storage::getInstance()->getDSMList();
    }
    else
    {
        dsms = dsm_list;
        dsms->push_back(m_dsid);
    }

    // loop over all dSMeter
    for (std::list<dsid_t>::iterator itm = dsms->begin();
         m_valid && (itm != dsms->end());
         itm++)
    {
        boost::shared_ptr<dSM> dsm;
        uint8_t aid;

        try
        {
            dsm = Storage::getInstance()->getDSM(*itm);
        }
        catch (SQLite3Exception &sx)
        {
            FILE_LOG(logERROR) << "Could not load dSM " << util::str(*itm) <<
                ": " << sx.getMessage();
            if (SQLite3::isFatal(sx.getErrorCode()))
            {
                throw sx;
            }
            continue;
        }
        catch (Exception &ex)
        {
            FILE_LOG(logERROR) << "[Task SetApartmentId] dSM with ID " <<
                    util::str(*itm) << " not found in database";
            continue;
        }

        dsa->dS485()->getApartmentId(*itm, &aid);

        DSID_LOG(logDEBUG2, *itm) << "[Task SetApartmentId] "
                "Database = " << util::str(m_apartment_id) <<
                ", Datamodel = " << util::str(dsm->getApartmentId()) <<
                ", Device = " << util::str(aid);

        if ((uint8_t) m_apartment_id != aid)
        {
            if (dsm->getApartmentId() >= 64)
            {
                FILE_LOG(logERROR) <<
                    "[Task SetApartmentId] Id mismatch on dsm " <<
                    util::str(*itm) <<
                    ": Database value " << util::str(m_apartment_id) <<
                    ", device value " << util::str(aid) <<
                    ", repeatedly failed to set id #" <<
                    util::str(dsm->getApartmentId() - 64);

                dsm->setApartmentId(dsm->getApartmentId() + 1);

                // TODO: Device Error Management
                // Populate misconfiguration error to dsm status monitor
                // CheckDSMTask
            }

            for (int retry = 0; retry < 5; retry ++)
            {
                DSID_LOG(logINFO, *itm) << "[Task SetApartmentId] Id "
                "reconfiguration from " << util::str(aid) <<
                " to " << util::str(m_apartment_id);

                aid = (uint8_t) m_apartment_id;
                try
                {
                    dsa->dS485()->setApartmentId(*itm, aid);
                }
                catch (Exception ex)
                {
                    DSID_LOG(logERROR, *itm) << "[Task SetApartmentId] "
                    "reconfiguration failed: " << ex.getMessage();
                    continue;
                }

                try
                {
                    dsa->dS485()->restartMeter(*itm);
                }
                catch(Exception ex)
                {
                    DSID_LOG(logERROR, *itm) << "[Task SetApartmentId] "
                    "restart failed: " << ex.getMessage();
                    continue;
                }
                break;
            }

            if (dsm->getApartmentId() >= 64)
            {
                dsm->setApartmentId(dsm->getApartmentId() + 1);
            }
            else {
                dsm->setApartmentId(64);
            }
            Storage::getInstance()->addDSM(dsm);
        }
    }
}

CheckDSMTask::CheckDSMTask(dsid_t dsid, bool async) : dSTask(), m_dsid(dsid),
        m_async(async)
{
    m_name = "dSM health check handler";
    m_description = "Check dSM state and configuration";
    m_valid = true;
}

void CheckDSMTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    // 1. Query dSMState() and set global health status accordingly
    {
        boost::shared_ptr<dSM> dsm;
        try
        {
            dsm = Storage::getInstance()->getDSM(m_dsid);
        }
        catch (SQLite3Exception &sx)
        {
            FILE_LOG(logERROR) << "Could not load dSM " << util::str(m_dsid) <<
                ": " << sx.getMessage();
            if (SQLite3::isFatal(sx.getErrorCode()))
            {
                throw sx;
            }
            return;
        }
        catch (Exception &ex)
        {
            FILE_LOG(logERROR) << "[Task CheckDSM] dSM with ID " <<
                    util::str(m_dsid) + " not found in database";
            return;
        }

        try
        {
            uint8_t Hardware, DSP, Temperature, Config, DS485, Line;
            dsa->dS485()->dSMState(m_dsid, &Hardware, &DSP, &Temperature, &Config, &DS485, &Line);
            DSID_LOG(logDEBUG, m_dsid) << "[Task CheckDSM] "
                    "Status HW:" << util::str(Hardware) << ", "
                    "DSP:" << util::str(DSP) << ", "
                    "Temperature:" << util::str(Temperature) << ", "
                    "Config:" << util::str(Config) << ", "
                    "DS485:" << util::str(DS485) << ", "
                    "Line:" << util::str(Line);
        }
        catch (Exception &ex)
        {
            DSID_LOG(logERROR, m_dsid) << "[Task CheckDSM] "
                    "Failed to query dsm status: " << ex.getMessage();
            throw ex;
        }

        bool restartRequired = false;
        try
        {
            uint8_t flags, newflags;
            bool led_on;
            bool progmode;
            ConfigManager::getInstance()->getParameter(
                    "digitalstrom.meter.led", led_on);
            ConfigManager::getInstance()->getParameter(
                    "digitalstrom.programmingmode", progmode);
            dsa->dS485()->getPropertyFlags(m_dsid, &flags);
            newflags = flags;

            uint8_t fvalue = led_on ? 1 : 0;
            if ((flags & 1) != fvalue)
            {
                newflags &= ~1;
                newflags |= fvalue;
                restartRequired = true;
            }

            fvalue = progmode ? 0 : 8;
            if ((flags & 8) != fvalue)
            {
                newflags &= ~8;
                newflags |= fvalue;
            }

            if (newflags != flags)
            {
                dsa->dS485()->setPropertyFlags(m_dsid, newflags);
                DSID_LOG(logINFO, m_dsid) << "[Task CheckDSM] "
                        "Reconfigure dSM Flags " <<
                        util::hex(flags, 2) << " => " << util::hex(newflags, 2);
            }
        }
        catch(Exception ex)
        {
            DSID_LOG(logERROR, m_dsid) << "[Task CheckDSM] "
                    "Reconfigure dSM Flags: " << ex.getMessage();
        }

        // Restart dSM if required, reset present flags
        try
        {
            if (restartRequired)
            {
                dsa->dS485()->restartMeter(m_dsid);
                dsm->setPresent(0);
            }
        }
        catch(Exception ex)
        {
            DSID_LOG(logERROR, m_dsid) << "[Task StartupDSM] "
                    "Restart dSM failed: " << ex.getMessage();
        }

        // TODO: status delivery and further processing?


    }
}

CheckDSMThread::CheckDSMThread(int interval) :
        dSLongTask(), m_interval(interval)
{
    m_name = "Periodic meter monitor";
    m_description = "Periodically check dSM state and configuration";
    m_valid = true;
    m_pausable = true;
    m_state = 0;
}

task_type CheckDSMThread::getTaskType() const
{
    return UnknownTask;
}

void CheckDSMThread::run(dS_Assistant *dsa)
{
    bool resetApartmentId = false;
    int newApartmentId = 0;

    if (!(dsa && m_valid))
    {
        m_status_mutex.lock();
        m_task_status = Cancelled;
        m_status_mutex.unlock();
        return;
    }

    if (m_paused)
    {
        m_paused = false;
    }

    m_status_mutex.lock();
    m_task_status = Running;
    m_status_mutex.unlock();

    // Automatic ApartmentId Reconfiguration
    {
        int apartmentid, busmembercount;
        dsid_t device_list[DSM_API_MAX_DEVICES];
        try
        {
            ConfigManager::getInstance()->getParameter("digitalstrom.apartmentid", apartmentid);
            busmembercount = dS_Assistant::getInstance()->dS485()->getBusMembers(
                    device_list, DSM_API_MAX_DEVICES);
            if (apartmentid + busmembercount >= 63)
            {
                newApartmentId = 63 - busmembercount;
                newApartmentId &= 0x38;  // 11 1000
                ConfigManager::getInstance()->setParameter("digitalstrom.apartmentid", newApartmentId);
                FILE_LOG(logWARNING) << "[Thread CheckDSM] "
                        "ApartmentId reconfiguration, new base address " <<
                        util::str(newApartmentId);
                resetApartmentId = true;
            }
        }
        catch (Exception ex)
        {
            FILE_LOG(logERROR) << "[Thread CheckDSM] "
                    "Failed to query apartment id settings: "
                    << ex.getMessage();
        }
    }

    while (m_valid)
    {
        if (m_paused)
        {
            m_status_mutex.lock();
            m_task_status = Paused;
            m_status_mutex.unlock();
            return;
        }

        switch (m_state)
        {
            case 0:
            {
                m_dsms = Storage::getInstance()->getDSMList();
                m_index = 0;
                m_state ++;
            }
            break;

            case 1:
            {
                if (m_dsms->empty())
                {
                    m_state ++;
                }
                else
                {
                    dsid_t dsmdsid = m_dsms->front();
                    m_dsms->pop_front();
                    m_index ++;

                    boost::shared_ptr<dSM> dsm;
                    try
                    {
                        dsm = Storage::getInstance()->getDSM(dsmdsid);
                    }
                    catch (SQLite3Exception &sx)
                    {
                        FILE_LOG(logERROR) << "Could not load dSM " <<
                            util::str(dsmdsid) << ": " << sx.getMessage();
                        if (SQLite3::isFatal(sx.getErrorCode()))
                        {
                            throw sx;
                        }
                    }
                    catch (Exception &ex)
                    {
                    }

                    if (dsm.get() && dsm->getPresent())
                    {
                        boost::shared_ptr<CheckDSMTask>
                            task(new CheckDSMTask(dsmdsid, false));
                        dsa->addTask(task);

                        if (resetApartmentId)
                        {
                            boost::shared_ptr<SetDSMApartmentIdTask>
                                task2(new SetDSMApartmentIdTask(dsmdsid,
                                        newApartmentId, false));
                            dsa->addTask(task2);
                        }
                    }
                }
            }
            break;

            default:
            {
                m_status_mutex.lock();
                m_task_status = Finished;
                m_status_mutex.unlock();
                if (m_interval > 0)
                {
                    boost::shared_ptr<CheckDSMThread> task(
                            new CheckDSMThread(m_interval));
                    dsa->scheduleLongTask(
                            boost::shared_ptr<dSLongTask> (task), m_interval);
                }
                return;
            }
            break;

        }
    }
}


GetActiveDevicesInZoneTask::GetActiveDevicesInZoneTask(dsid_t dsid,
                                                       uint16_t zone_id) :
                                                       dSTask(), m_dsid(dsid),
                                                       m_zone_id(zone_id)
{
    m_name = "Getting active devices from zone";
    m_description = "Retrieve information about active devices in a given zone"
                    " on a dS-Meter";
    m_valid = true;
}

void GetActiveDevicesInZoneTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    uint16_t count;

    uint16_t device_id;
    uint16_t vendor_id;
    uint16_t product_id;
    uint16_t function_id;
    uint16_t version;
    uint16_t zone_id;
    uint8_t state;
    uint8_t locked;
    uint8_t output_mode;
    uint8_t name[21];
    uint32_t serial_number;

    dsa->dS485()->getActiveDevicesCountInZone(m_dsid, m_zone_id, &count);
    for (uint16_t i = 0; i < count; i++)
    {
        dsa->dS485()->getActiveDeviceInfoInZone(m_dsid, m_zone_id, i,
                                       &device_id,
                                       &vendor_id, &product_id, &function_id,
                                       &version, &zone_id, &state, NULL,
                                       &locked, &output_mode, NULL,
                                       name, NULL,
                                       NULL, &serial_number, NULL);

        dsid_t device_dsid;
        if (DsmApiExpandDeviceDSID(vendor_id, serial_number, &device_dsid) != 0)
        {
            FILE_LOG(logERROR) <<
                "[Task GetActiveDevices] Could not resolve dsid: "
                "serial " << util::hex(device_id) <<
                "/vendor " << util::hex(vendor_id);
            continue;
        }

        boost::shared_ptr<Device> dev(new Device(device_dsid, m_dsid));
        dev->setId(device_id);
        dev->setProductId(product_id);
        dev->setFunctionId(function_id);
        dev->setVersion(version);
        dev->setZoneId(zone_id);
        dev->setState(state);
        dev->setLocked(locked);
        dev->setOutputMode(output_mode);
        dev->setName((const char*)name);
        dev->setSerialNumber(serial_number);
        Storage::getInstance()->addDevice(dev);

        FILE_LOG(logINFO) << "Added " << (state ? "active" : "inactive")
                          << " device with dSID " << util::str(dev->getDSID())
                          << " and address " << util::str(device_id)
                          << " on dSM " << util::str(dev->getDSMdSID());
    }
}

GetAllDevicesTask::GetAllDevicesTask(dsid_t dsid) :
        dSTask(), m_dsid(dsid)
{
    m_name = "Getting all devices on a circuit";
    m_description = "Retrieve information about all devices on a given circuit";
    m_valid = true;
}

void GetAllDevicesTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    uint16_t count;

    uint16_t device_id;
    uint16_t vendor_id;
    uint16_t product_id;
    uint16_t function_id;
    uint16_t version;
    uint16_t zone_id;
    uint8_t state;
    uint8_t locked;
    uint8_t output_mode;
    uint8_t name[21];
    uint32_t serial_number;

    dsa->dS485()->getDevicesCountInZone(m_dsid, 0, &count);
    for (uint16_t i = 0; i < count; i++)
    {
        try {
            dsa->dS485()->getDeviceInfoByIndex(m_dsid, 0, i,
                    &device_id, &vendor_id, &product_id, &function_id,
                    &version, &zone_id, &state, NULL,
                    &locked, &output_mode, NULL,
                    name, NULL,
                    NULL, &serial_number, NULL);
        }
        catch (Exception& ex)
        {
            FILE_LOG(logERROR) << "dsm-api call exception: dSM:"
                               << util::str(m_dsid) << ", device index "
                               << util::str(i);
            continue;
        }

        dsid_t device_dsid;
        if (DsmApiExpandDeviceDSID(vendor_id, serial_number, &device_dsid) != 0)
        {
            FILE_LOG(logERROR) <<
                    "[Task GetAllDevices] Could not resolve dsid: "
                    "serial " << util::hex(device_id) <<
                    "/vendor " << util::hex(vendor_id);
            continue;
        }

        boost::shared_ptr<Device> dev(new Device(device_dsid, m_dsid));
        dev->setId(device_id);
        dev->setProductId(product_id);
        dev->setFunctionId(function_id);
        dev->setVersion(version);
        dev->setZoneId(zone_id);
        dev->setState(state);
        dev->setLocked(locked);
        dev->setOutputMode(output_mode);
        dev->setName((const char*)name);
        dev->setSerialNumber(serial_number);

        //TODO: set zone index

        Storage::getInstance()->addDevice(dev);

        FILE_LOG(logINFO) << "Added " << (state ? "active" : "inactive")  <<
                " device with dSID " << util::str(dev->getDSID()) <<
                " and address " << util::str(device_id) <<
                " on dSM " << util::str(dev->getDSMdSID());
    }
}

GetDSMZonesTask::GetDSMZonesTask(dsid_t dsid) : dSTask(), m_dsid(dsid)
{
    m_name = "Getting number of zones on dSM";
    m_description = "Retrieve number of zones for dSM " + util::str(dsid);
    m_valid = true;
}

void GetDSMZonesTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    uint8_t count;
    dsa->dS485()->getZoneCount(m_dsid, &count);
    for (uint8_t i = 0; i < count; i++)
    {
        boost::shared_ptr<GetDSMZoneInfoTask> task(
                                            new GetDSMZoneInfoTask(m_dsid, i));
        dsa->addTask(boost::dynamic_pointer_cast<dSTask>(task));
    }
}

GetDSMZoneInfoTask::GetDSMZoneInfoTask(dsid_t dsid, uint8_t zone_index) :
                                       dSTask(), m_dsid(dsid),
                                       m_zone_index(zone_index)
{
    m_name = "Getting zone information";
    m_description = "Retrieve information for zone with index " +
                     util::str(zone_index) + " on dSM " + util::str(dsid);
    m_valid = true;
}

void GetDSMZoneInfoTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    uint16_t zone_id;
    uint8_t vzone_id;
    uint8_t num_grp;
    uint8_t name[21];

    dsa->dS485()->getZoneInfo(m_dsid, m_zone_index, &zone_id, &vzone_id,
                              &num_grp, name);
    FILE_LOG(logINFO) << "Retrieved zone info for dsm: "
        << util::str(m_dsid) << ", zoneid: " << (int)zone_id << " vzid: "
        << (int)vzone_id << " groups:  " << (int)num_grp << "name: " << name;
}

GetDeviceInfoTask::GetDeviceInfoTask(dsid_t dsm_dsid, uint16_t device_id) :
                                     dSTask(), m_dsid(dsm_dsid),
                                     m_device_id(device_id)
{
    m_name = "Getting device information";
    m_description = "Retrieve device information for device on dSM " +
                   util::str(dsm_dsid);
    m_valid = true;
}

void GetDeviceInfoTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    uint16_t device_id;
    uint16_t vendor_id;
    uint16_t product_id;
    uint16_t function_id;
    uint16_t version;
    uint16_t zone_id;
    uint8_t state;
    uint8_t locked;
    uint8_t output_mode;
    uint8_t name[21];
    uint32_t serial_number;

    dsa->dS485()->getDeviceInfoById(m_dsid, m_device_id,
                       &device_id, &vendor_id, &product_id, &function_id,
                       &version, &zone_id, &state, NULL, &locked,
                       &output_mode, NULL, name, NULL,
                       NULL, &serial_number, NULL);

    dsid_t device_dsid;
    if (DsmApiExpandDeviceDSID(vendor_id, serial_number, &device_dsid) != 0)
    {
        FILE_LOG(logWARNING) << "Could not resolve dsid: vendor id is not "
            << "known, ignoring device";
        return;
    }

    boost::shared_ptr<Device> dev;

    try
    {
        dev = Storage::getInstance()->getDevice(device_dsid, m_dsid);
    }
    catch (SQLite3Exception sx)
    {
        FILE_LOG(logERROR) << "Could not load device " <<
            util::str(device_dsid) << ": " << sx.getMessage();
        if (SQLite3::isFatal(sx.getErrorCode()))
        {
            throw sx;
        }
        dev = boost::shared_ptr<Device>(new Device(device_dsid, m_dsid));
    }
    catch (Exception ex)
    {
        dev = boost::shared_ptr<Device>(new Device(device_dsid, m_dsid));
    }

    dev->setId(device_id);
    dev->setProductId(product_id);
    dev->setFunctionId(function_id);
    dev->setVersion(version);
    dev->setZoneId(zone_id);
    dev->setState(state);
    dev->setLocked(locked);
    dev->setOutputMode(output_mode);
    dev->setName((const char*)name);
    Storage::getInstance()->addDevice(dev);

    FILE_LOG(logINFO) << "Updated device info with dSID "
                      << util::str(dev->getDSID()) <<  " on dSM "
                      << util::str(dev->getDSMdSID());
}

TestTranmissionQualityTask::TestTranmissionQualityTask(int checksPerDevice,
                                dsid_t dsm_dsid) : dSLongTask(),
                                m_checksPerDevice(checksPerDevice),
                                m_dsm_dsid(dsm_dsid)
{
    m_name = "Test der Übertragungsqualität der Klemmen";
    m_description = "Testen der Übertragungsqualität für ";
    m_parameters = "dsm=" + util::str(dsm_dsid) + '&' +
                   "checksPerDevice=" + util::str(checksPerDevice);

    if (IsNullId(dsm_dsid))
    {
        m_description += "alle Klemmen";
    }
    else
    {
        m_description += "alle Klemmen am dSM " + util::str(dsm_dsid);
    }

    m_description += ", Tests pro Klemme: " + util::str(checksPerDevice);

    m_valid = true;
    m_pausable = true;
    m_state = kStateInit;
}

TestTranmissionQualityTask::~TestTranmissionQualityTask()
{
}

task_type TestTranmissionQualityTask::getTaskType() const
{
    return TestTransmissionQuality;
}

void TestTranmissionQualityTask::callback(boost::shared_ptr<CBData> busdata)
{
    switch(busdata->getType())
    {
        default:
            break;

        case DeviceTestTransmissionQuality:
        {
            boost::shared_ptr<DeviceTestQualityData> data =
                boost::dynamic_pointer_cast<DeviceTestQualityData> (busdata);
            m_result_up = data->getUpstream();
            m_result_down = data->getDownstream();
            m_result_err = 0;
        }
        break;
    }
}

void TestTranmissionQualityTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        m_status_mutex.lock();
        m_task_status = Cancelled;
        m_status_mutex.unlock();
        return;
    }

    m_status_mutex.lock();
    m_task_status = Running;
    m_status_mutex.unlock();

    if (kStateInit == m_state)
    {
        if (IsNullId(m_dsm_dsid))
        {
            m_dev_list = Storage::getInstance()->getDevices(true);
        }
        else
        {
            m_dev_list = Storage::getInstance()->getDevices(m_dsm_dsid, true);
        }
        m_dev_count = 0;
        m_dev_current = m_dev_list->begin();
        m_dev_total = m_dev_list->size() * m_checksPerDevice;
        m_check_loop = 0;
        m_progress = 0;

        dsa->dS485()->addListener(this);
        this->select(DeviceTestTransmissionQuality, true);

        m_state = kStateSetup;
    }

    if (m_paused)
    {
        m_paused = false;
    }

    if (0 == m_dev_total)
    {
        m_status_mutex.lock();
        m_progress = 100;
        m_task_status = Finished;
        m_status_mutex.unlock();
        return;
    }

    while (m_valid && (m_dev_count <= m_dev_total) && (m_state != kStateFinished))
    {

        if (m_paused)
        {
            m_status_mutex.lock();
            m_task_status = Sleeping;
            m_status_mutex.unlock();
            return;
        }

        switch (m_state)
        {
            case kStateSetup:
            {
                m_result_down = TRANSMISSION_QUALITY_DOWNSTREAM_MAX;
                m_result_up = 0;
                m_result_err = ERROR_TIMEOUT;

                boost::shared_ptr<Device> device = *m_dev_current;
                m_dev_count ++;

                m_status_mutex.lock();
                m_status_info = "Requesting transmission quality from device " +
                        util::str(device->getDSID());
                if (m_checksPerDevice > 1)
                {
                    m_status_info += ", test " + util::str(m_check_loop + 1);
                }
                m_status_mutex.unlock();

                try
                {
                    dsa->dS485()->reqDeviceTransmissionQuality(device->getDSMdSID(), device->getId());
                    m_state = kStateRequest;
                    dsa->scheduleLongTask(
                            boost::shared_ptr<dSLongTask> (shared_from_this()),
                            2);
                    return;
                }
                catch (Exception ex)
                {
                    m_status_mutex.lock();
                    m_status_info = "Failed to get transmission quality for "
                            "device " + util::str(device->getDSID()) +
                            ": " + ex.getMessage();
                    m_result_err = ex.getErrorCode();
                    m_status_mutex.unlock();

                    FILE_LOG(logINFO) << "[Task Transmission Quality] Device " <<
                            util::str(device->getDSID()) <<
                            ", error " << util::str(m_result_err) << " = "
                            << ex.getMessage();
                    m_state = kStateResponse;
                }
            }
            break;

            case kStateRequest:
            {
                m_state = kStateResponse;
                if (m_result_err == ERROR_TIMEOUT)
                {
                    dsa->scheduleLongTask(
                            boost::shared_ptr<dSLongTask> (shared_from_this()),
                            2);
                    return;
                }
            }
            break;

            case kStateResponse:
            {
                m_state = kStateNext;
                boost::shared_ptr<Device> device = *m_dev_current;
                Database::getInstance()->saveTransmissionQuality(m_task_id,
                        util::str(device->getDSID()),
                        m_result_up, m_result_down, m_result_err);

                m_status_mutex.lock();
                m_progress = (unsigned char)(m_dev_count * 100 / m_dev_total);
                m_status_mutex.unlock();

                FILE_LOG(logDEBUG1) << "[Task Transmission Quality] "
                        "Progress " << util::str(m_progress) <<
                        ", Device " << util::str(device->getDSID()) <<
                        ", down " << util::str(m_result_down) <<
                        ", up " << util::str(m_result_up);
            }
            break;

            case kStateNext:
            {
                m_state = kStateSetup;
                m_dev_current ++;
                if (m_dev_current == m_dev_list->end())
                {
                    m_check_loop++;
                    if (m_check_loop < m_checksPerDevice)
                    {
                        m_dev_current = m_dev_list->begin();
                    }
                    else
                    {
                        m_state = kStateFinished;
                    }
                }
            }
            break;

            case kStateFinished:
            {
                m_dev_count = m_dev_total;
                FILE_LOG(logWARNING) << "[Task Transmission Quality] Force task end";
            }
            break;

            default:
                m_state = kStateFinished;
                break;

        } // switch
    }

    FILE_LOG(logINFO) << "[Task Transmission Quality] Ready, sent " <<
                                util::str(m_dev_count) << " requests";

    dsa->dS485()->removeListener(this);

    m_status_mutex.lock();
    if (!m_valid)
    {
        m_task_status = Cancelled;
        m_status_info = "Cancelled";
    }
    else
    {
        m_task_status = Finished;
        m_status_info = "Finished";
    }
    m_status_mutex.unlock();
}


TestOneDevicePresenceTask::TestOneDevicePresenceTask(dsid_t dsid,
        dsid_t dsm_dsid, int timeout) :
        dSLongTask(),
        m_dsid(dsid),
        m_dsm_dsid(dsm_dsid),
        m_timeout(timeout)
{
    m_name = "Geräteanwesenheit";
    m_description = "Aktualisiere Anwesenheit des Gerätes " + util::str(m_dsid);
    m_parameters = "dsid=" + util::str(dsid) + "&dsm_dsid=" + util::str(dsm_dsid);
    m_valid = true;
    m_pausable = true;
    m_state = kStateInit;
    m_present = false;
}

TestOneDevicePresenceTask::~TestOneDevicePresenceTask()
{
}

task_type TestOneDevicePresenceTask::getTaskType() const
{
    return TestOneDevicePresence;
}

void TestOneDevicePresenceTask::callback(boost::shared_ptr<CBData> busdata)
{
    switch(busdata->getType())
    {
        default:
            break;

        case DevicePing:
        {
            boost::shared_ptr<DevicePingData> data =
                    boost::dynamic_pointer_cast<DevicePingData> (busdata);
            m_present = data->getPresent();
            m_result_err = 0;
        }
        break;
    }
}

void TestOneDevicePresenceTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        m_status_mutex.lock();
        m_task_status = Cancelled;
        m_status_mutex.unlock();
        return;
    }

    m_status_mutex.lock();
    m_task_status = Running;
    m_status_mutex.unlock();

    if (kStateInit == m_state)
    {
        m_progress = 0;
        m_waiting = 0;
        try {
            m_device = Storage::getInstance()->getDevice(m_dsid, m_dsm_dsid);
        }
        catch (SQLite3Exception sx)
        {
            FILE_LOG(logERROR) << "Could not load device: " << sx.getMessage();
            if (SQLite3::isFatal(sx.getErrorCode()))
            {
                FILE_LOG(logERROR) << "Unrecoverable SQLite error, terminating dSA";
                g_shutdown_flag = SHUTDOWN_SQLITE;
            }
        }
        catch (Exception ex)
        {
            m_status_mutex.lock();
            m_task_status = Finished;
            m_status_info = "Unknown device with DSID " + util::str(m_dsid);
            m_result_err = ex.getErrorCode();
            m_status_mutex.unlock();
            m_progress = 100;
            return;
        }

        dsa->dS485()->addListener(this);
        this->select(DevicePing, true);

        m_state = kStateSetup;
    }

    if (m_paused)
    {
        m_paused = false;
    }

    while (m_valid && (m_state != kStateFinished))
    {

        if (m_paused)
        {
            m_status_mutex.lock();
            m_task_status = Sleeping;
            m_status_mutex.unlock();
            return;
        }

        switch (m_state)
        {
            case kStateSetup:
            {
                FILE_LOG(logINFO) << "[Task OneDevice Presence] "
                        "Device " << util::str(m_dsid);

                try
                {
                    m_result_err = ERROR_TIMEOUT;
                    dsa->dS485()->reqDevicePresenceTest(m_dsm_dsid, m_device->getId());
                    m_state = kStateRequest;
                    dsa->scheduleLongTask(
                            boost::shared_ptr<dSLongTask> (shared_from_this()),
                            2);
                    return;
                }
                catch (Exception ex)
                {
                    m_status_mutex.lock();
                    m_status_info = "Failed to update presence for "
                            "device " + util::str(m_dsid) +
                            ": " + ex.getMessage();
                    m_result_err = ex.getErrorCode();
                    m_status_mutex.unlock();

                    FILE_LOG(logINFO) << "[Task OneDevice Presence] Device " <<
                            util::str(m_dsid) <<
                            ", error " << util::str(m_result_err) << " = "
                            << ex.getMessage();
                    m_state = kStateResponse;
                }
            }
            break;

            case kStateRequest:
            {
                if ((m_result_err == ERROR_TIMEOUT) && (m_waiting < m_timeout))
                {
                    m_waiting += 2;
                    dsa->scheduleLongTask(
                            boost::shared_ptr<dSLongTask> (shared_from_this()),
                            2);
                    m_status_info = "Warte auf Antwort von Gerät "
                                                           + util::str(m_dsid);
                    return;
                }
                m_state = kStateResponse;
            }
            break;

            case kStateResponse:
            {
                m_state = kStateFinished;
                m_status_mutex.lock();
                if (m_result_err == ERROR_TIMEOUT)
                {
                    m_status_info = "Keine Antwort von Gerät "
                                                           + util::str(m_dsid);
                }
                else if (m_result_err == ERROR_OK)
                {
                    m_status_info = "Antwort erhalten von Gerät "
                                                           + util::str(m_dsid);
                }
                else
                {
                    m_status_info = "Fehler bei der Anfrage an Gerät "
                          + util::str(m_dsid) + ": " + util::str(m_result_err);
                }
                m_progress = 100;
                m_status_mutex.unlock();

                FILE_LOG(logDEBUG1) << "[Task OneDevice Presence] "
                        "Device " << util::str(m_dsid) <<
                        " present: " << (m_present ? "yes" : "no");
            }
            break;

            case kStateFinished:
            {
                FILE_LOG(logWARNING) << "[Task OneDevice Presence] Force task end";
            }
            break;

            default:
            {
                m_state = kStateFinished;
            }
            break;

        } // switch
    }

    FILE_LOG(logINFO) << "[Task OneDevice Presence] Ready";

    dsa->dS485()->removeListener(this);

    m_status_mutex.lock();
    if (!m_valid)
    {
        m_task_status = Cancelled;
    }
    else
    {
        m_task_status = Finished;
    }
    m_status_mutex.unlock();
}

TestDevicePresenceTask::TestDevicePresenceTask(dsid_t dsm_dsid,
                                               int interval_time) :
        dSLongTask(),
        m_dsm_dsid(dsm_dsid),
        m_interval_time(interval_time),
        m_devs(new std::list<boost::shared_ptr<DeviceP> >),
        m_dsms(new std::list<boost::shared_ptr<MeterP> >)
{
    m_name = "Geräteanwesenheitstest";
    m_description = "Prüfe Anwesenheit registrierter Geräte ";
    m_parameters = "dsm=" + util::str(dsm_dsid) + '&' +
            "interval=" + util::str(interval_time);

    if (! IsNullId(dsm_dsid))
    {
        m_description += "am dSM " + util::str(dsm_dsid);
    }
    else
    {
        m_description += "aller dSM's";
    }
    m_valid = true;
    m_pausable = true;
    m_state = 0;

    pthread_mutex_init(&m_ping_lock, NULL);
    pthread_cond_init(&m_ping_status_change, NULL);
}

TestDevicePresenceTask::~TestDevicePresenceTask()
{
    pthread_mutex_destroy(&m_ping_lock);
    pthread_cond_destroy(&m_ping_status_change);
}

task_type TestDevicePresenceTask::getTaskType() const
{
    return TestDevicePresence;
}

TestDevicePresenceTask::DeviceP::DeviceP(Device* parent) :
        Device(*parent)
{
    m_count_ping_sent = m_count_ping_received = 0;
}

TestDevicePresenceTask::MeterP::MeterP(dSM* parent, TestDevicePresenceTask* const task) :
        dSM(*parent), m_task(task)
{
    resetPingState();
}

void TestDevicePresenceTask::MeterP::resetPingState()
{
    m_count_ping_dut = m_count_ping_sent = m_count_ping_received = m_deviceid_dut = 0;
    m_state = pm_state_unknown;
}

void TestDevicePresenceTask::MeterP::getPingStatus(uint16_t* tx_dev_count,
                                                   uint16_t* tx_count,
                                                   uint16_t* rx_count)
{
    *tx_dev_count = m_count_ping_dut;
    *tx_count = m_count_ping_sent;
    *rx_count = m_count_ping_received;
}

void TestDevicePresenceTask::MeterP::signalPingSent(uint16_t device_id)
{
    if (device_id < 64)
    {
        return;
    }
    if (device_id != m_deviceid_dut)
    {
        m_deviceid_dut = device_id;
        m_count_ping_dut ++;
        m_count_ping_timeout = 0;

        boost::shared_ptr<DeviceP> dsd = m_task->getDevice(getDSID(), device_id);
        if (dsd)
        {
            Database::getInstance()->updateDevicePresenceStatus(m_task->m_task_id,
                    util::str(dsd->getDSMdSID()), device_id,
                    dsd->getPingState());
        }
    }
    m_state = pm_state_ping_sent;
    m_count_ping_sent ++;
    pthread_cond_signal(&m_task->m_ping_status_change);
}

void TestDevicePresenceTask::MeterP::signalPingReply(uint16_t device_id)
{
    if (device_id == m_deviceid_dut)
    {
        m_count_ping_received ++;
        m_state = pm_state_reply_received;

        boost::shared_ptr<DeviceP> dsd = m_task->getDevice(getDSID(), device_id);
        if (dsd)
        {
            Database::getInstance()->updateDevicePresenceStatus(m_task->m_task_id,
                    util::str(dsd->getDSMdSID()), device_id,
                    dsd->getPingState());
        }

        if (m_count_ping_dut >= getDeviceCount())
        {
            m_state = pm_state_ready;
        }
    }
    pthread_cond_signal(&m_task->m_ping_status_change);
}

void TestDevicePresenceTask::MeterP::signalPingTimeout()
{
    m_state = pm_state_waiting;
    if (++m_count_ping_timeout >= 3)
    {
        if (m_count_ping_dut >= getDeviceCount())
        {
            m_state = pm_state_ready;
        }
    }
}

boost::shared_ptr<TestDevicePresenceTask::MeterP>
                                    TestDevicePresenceTask::getDSM(dsid_t dsmid)
{
    dsid_t id = m_dsm_under_test->getDSID();
    if (IsEqualId(id, dsmid))
    {
        return m_dsm_under_test;
    }
    return boost::shared_ptr<TestDevicePresenceTask::MeterP>();
}

boost::shared_ptr<TestDevicePresenceTask::DeviceP>
            TestDevicePresenceTask::getDevice(dsid_t dsmid, uint16_t devindex)
{
    for (std::list<boost::shared_ptr<
            TestDevicePresenceTask::DeviceP> >::iterator itm = m_devs->begin();
            (itm != m_devs->end());
            itm++)
    {
        dsid_t idm = (*itm)->getDSMdSID();
        if (IsEqualId(idm, dsmid) && ((*itm)->getId() == devindex))
        {
            return *itm;
        }
    }
    return boost::shared_ptr<TestDevicePresenceTask::DeviceP>();
}

boost::shared_ptr<TestDevicePresenceTask::DeviceP>
                TestDevicePresenceTask::getDevice(dsid_t dsmid, dsid_t devid)
{
    for (std::list<boost::shared_ptr<
            TestDevicePresenceTask::DeviceP> >::iterator itm = m_devs->begin();
            (itm != m_devs->end());
            itm++)
    {
        dsid_t idm = (itm->get())->getDSMdSID();
        dsid_t idd = (itm->get())->getDSID();
        if (IsEqualId(idm, dsmid) && IsEqualId(idd, devid))
        {
            return *itm;
        }
    }
    return boost::shared_ptr<TestDevicePresenceTask::DeviceP>();
}

void TestDevicePresenceTask::callback(boost::shared_ptr<CBData> busdata)
{
    switch(busdata->getType())
    {

        default:
            break;

        case CommUpstreamShort:
            break;

        case CommUpstreamLong:
        {
            boost::shared_ptr<CommUpstreamLongData> data =
                boost::dynamic_pointer_cast<CommUpstreamLongData> (busdata);
            if (data->getCrosstalkFlags() > 0)
            {
                return;
            }
            switch (data->getMainQualifier())
            {
                case 15:
                {
                    // NOTE: this 32bit -> DSID conversion assumes
                    // vendor id 0x0001
                    dsid_t device_dsid;
                    DsmApiExpandDeviceDSID(0x0001, data->getData(),
                            &device_dsid);
                    boost::shared_ptr<MeterP> dsm =
                            getDSM(data->getDSMSourceDSID());
                    boost::shared_ptr<DeviceP> dsd =
                            getDevice(data->getDSMSourceDSID(), device_dsid);

                    std::string sourceId((dsd ? util::str(dsd->getId()) : "0000") +
                            "/" + util::hex(data->getData()));

                    FILE_LOG(logDEBUG2) << "Ping reply from " << sourceId;

                    if (dsd && dsd->getPingState() == pd_state_sent)
                    {
                        switch(dsd->getPingState())
                        {
                            case pd_state_sent:
                                dsd->setPingState(pd_state_replied);
                                if (dsm)
                                {
                                    dsm->signalPingReply(dsd->getId());
                                }
                                break;
                            case pd_state_replied:
                                break;
                            default:
                                break;
                        }
                    }
                }
                break;
            }
        }
        break;

        case CommDownstreamShort:
        break;

        case CommDownstreamLong0:
        {
            boost::shared_ptr<CommDownstreamLong0Data> data(
                boost::dynamic_pointer_cast<CommDownstreamLong0Data> (busdata));
            switch (data->getCommand())
            {
                case 0x02:
                {
                    boost::shared_ptr<MeterP> dsm =
                            getDSM(data->getDSMSourceDSID());
                    boost::shared_ptr<DeviceP> dsd =
                            getDevice(data->getDSMSourceDSID(), dsm->getSelectionB() >> 6);

                    std::string targetId(util::str(dsm->getSelectionB() >> 6) + "/" +
                            (dsd ? util::str(dsd->getDSID()) : "00000000"));

                    if (dsm)
                    {
                        FILE_LOG(logDEBUG2) << "Ping sent to " << targetId;
                        dsm->signalPingSent(dsm->getSelectionB() >> 6);
                    }
                    if (dsd)
                    {
                        if (dsd->getPingState() == pd_state_scheduled)
                        {
                            dsd->setPingState(pd_state_sent);
                        }
                        else if (dsd->getPingState() == pd_state_sent)
                        {
                        }
                    }
                } /* case 0x02 */
                break;
            } /* switch getCommand() */
        }
        break;

        case CommDownstreamLong8:
        break;

        case CommDownstreamLong16:
        {
            boost::shared_ptr<CommDownstreamLong16Data> data(
                boost::dynamic_pointer_cast<CommDownstreamLong16Data>(busdata));
            boost::shared_ptr<MeterP> dsm = getDSM(data->getDSMSourceDSID());
            if (dsm)
            {
                static uint16_t dsidh = 0;
                uint16_t payload = data->getData();
                switch (data->getCommand())
                {
                    case 0:
                        dsm->setSelectionA(payload);
                        break;
                    case 1:
                        dsm->setSelectionB(payload);
                        dsm->setSelectionD(0);
                        break;
                    case 2:
                        dsidh = payload;
                        break;
                    case 3:
                        dsm->setSelectionD(
                                (uint32_t) dsidh << 16 | data->getData());
                        dsidh = 0;
                        break;
                }
            }
        }
        break;

        case DeviceAdded:
        {
            boost::shared_ptr<DeviceAddedData> data(
                    boost::dynamic_pointer_cast<DeviceAddedData> (busdata));
            boost::shared_ptr<DeviceP> dsd = getDevice(data->getDSMSourceDSID(),
                                                       data->getDeviceID());
            if (dsd)
            {
                dsd->setPingState(pd_state_discovered);
            }
        }
        break;

        case DeviceRemoved:
        {
            boost::shared_ptr<DeviceRemovedData> data(
                    boost::dynamic_pointer_cast<DeviceRemovedData> (busdata));
            boost::shared_ptr<DeviceP> dsd = getDevice(data->getDSMSourceDSID(),
                                                       data->getDeviceID());
            if (dsd)
            {
                dsd->setPingState(pd_state_removed);
            }
        }
        break;
    }
}

void TestDevicePresenceTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        m_status_mutex.lock();
        m_task_status = Cancelled;
        m_status_mutex.unlock();
        return;
    }

    m_status_mutex.lock();
    m_task_status = Running;
    m_progress = 0;
    m_status_mutex.unlock();

    // initial setup
    if (m_state == 0)
    {
        boost::shared_ptr<std::list<boost::shared_ptr<Device> > > devs;
        boost::shared_ptr<std::list<dsid_t> > dsms;
        boost::shared_ptr<std::list<dsid_t> > dsm_list(new std::list<dsid_t>);

        if (IsNullId(m_dsm_dsid))
        {
            devs = Storage::getInstance()->getDevices(false);
            dsms = Storage::getInstance()->getDSMList();
        }
        else
        {
            devs = Storage::getInstance()->getDevices(m_dsm_dsid, false);
            dsms = dsm_list;
            dsms->push_back(m_dsm_dsid);
        }

        // copy device list and preset device ping state
        for (std::list<boost::shared_ptr<Device> >::iterator it =
                devs->begin();
                it != devs->end();
                it++)
        {
            boost::shared_ptr<DeviceP> dsd(new DeviceP((*it).get()));
            dsd->setPingState(pd_state_scheduled);
            m_devs->push_back(dsd);

            Database::getInstance()->saveDevicePresence(m_task_id,
                            util::str(dsd->getDSID()),
                            util::str(dsd->getDSMdSID()),
                            dsd->getId(),
                            dsd->getPingState());
        }

        // copy meter list and reset ping state
        m_total_dev_count = 0;
        for (std::list<dsid_t>::iterator itm = dsms->begin();
                m_valid && (itm != dsms->end());
                itm++)
        {
            try
            {
                boost::shared_ptr<MeterP> dsm(
                        new MeterP(Storage::getInstance()->getDSM(*itm).get(), this));
                dsm->resetPingState();
                m_total_dev_count += (dsm->getActiveDeviceCount() + dsm->getInactiveDeviceCount());
                m_dsms->push_back(dsm);
            }
            catch (SQLite3Exception &sx)
            {
                FILE_LOG(logERROR) << "Could not load dSM " <<
                    util::str(*itm) << ": " << sx.getMessage();
                if (SQLite3::isFatal(sx.getErrorCode()))
                {
                    throw sx;
                }
            }
            catch (Exception &ex)
            {
                FILE_LOG(logERROR) << "[Task Device Presence] dSM with ID " <<
                        util::str(*itm) << " not found in database";
            }
        }

        if (0 == m_total_dev_count)
        {
            m_status_mutex.lock();
            m_progress = 100;
            m_task_status = Finished;
            m_status_mutex.unlock();
            return;
        }

        // disable all active pings
        dsid_t bid;
        SetBroadcastId(bid);
        try
        {
            dsa->dS485()->pingOff(bid);
        }
        catch (Exception &ex)
        {
            // Broadcast command -> result check raises exception for return code 2 :(
        }

        // add task to receive callbacks
        dsa->dS485()->addListener(this);

        // select the callbacks you want
        this->select(CommUpstreamShort, true);
        this->select(CommUpstreamLong, true);
        this->select(CommDownstreamShort, true);
        this->select(CommDownstreamLong0, true);
        this->select(CommDownstreamLong8, true);
        this->select(CommDownstreamLong16, true);
        this->select(DeviceAdded, true);
        this->select(DeviceRemoved, true);

        m_total_rx = m_total_tx = 0;
        m_tx_dev_count = m_tx_count = m_rx_count = 0;

        m_state ++;

        FILE_LOG(logINFO) << "[Task Device Presence] Starting presence check on " <<
                util::str(m_dsms->size()) << " meter(s) for a total of " <<
                util::str(m_total_dev_count) << " devices (db " <<
                util::str(m_devs->size()) << ")";
    }

    // resuming from paused state?
    if (m_paused)
    {
        m_paused = false;
    }

    while (m_valid && (m_state != 6))
    {
        if (m_paused)
        {
            m_status_mutex.lock();
            m_task_status = Sleeping;
            m_status_mutex.unlock();
            return;
        }

        switch (m_state)
        {
            case 1:
            {
                if (m_dsms->empty())
                {
                    m_state = 6;
                }
                else
                {
                    m_dsm_under_test = m_dsms->front();
                    m_dsms->pop_front();
                    m_dsm_dev_count = m_dsm_under_test->getDeviceCount();
                    if (m_dsm_dev_count > 0)
                    {
                        m_status_mutex.lock();
                        m_status_info = "Check " + util::str(m_dsm_dev_count) + " devices on dsm " +
                                util::str(m_dsm_under_test->getDSID());
                        m_status_mutex.unlock();
                        m_state ++;
                    }
                    m_rx_count = m_tx_count = 0;
                    m_tx_dev_count = 0;
                    m_tx_count_equal = 0;
                }
            }
            break;

            case 2:
            {
                uint8_t Hardware;
                uint8_t DSP;
                uint8_t Temperature;
                uint8_t Config;
                uint8_t dS485;
                uint8_t Line;
                try
                {
                    dS_Assistant::getInstance()->dS485()->dSMState(m_dsm_under_test->getDSID(),
                            &Hardware, &DSP, &Temperature, &Config, &dS485, &Line);
                    if (Line)
                    {
                        FILE_LOG(logINFO) << "[Task Device Presence] dSM " <<
                                util::str(m_dsm_under_test->getDSID()) << " is currently jumbling "
                                "and busy, device presence check skipped";
                        m_total_tx += m_dsm_dev_count;
                        m_state = 1;
                    }
                    else
                    {
                        m_state ++;
                    }
                }
                catch (Exception &ex)
                {
                    m_state = 1;
                }
            }
            break;

            case 3:
            {
                FILE_LOG(logDEBUG1) << "[Task Device Presence] Start to check " <<
                        util::str(m_dsm_dev_count) << " devices on dsm " <<
                        util::str(m_dsm_under_test->getDSID());

                // send ping-one-time-off command, after that the 24h mode is re-enabled
                try
                {
                    dsa->dS485()->pingOneTimeOff(m_dsm_under_test->getDSID(),
                            (uint16_t) m_interval_time);
                    m_state ++;
                }
                catch(Exception &ex)
                {
                    FILE_LOG(logERROR)
                            << "[Task Device Presence] dSM-API request failed: "
                            << ex.getMessage();
                    m_state = 1;
                }
            }
            break;

            case 4:
            {
                const uint16_t reschedule_time = 2;
                uint16_t tx_count_last = m_tx_count;
                uint16_t tx_threshold = (m_interval_time + 2500) / 1000;

                m_dsm_under_test->getPingStatus(&m_tx_dev_count, &m_tx_count, &m_rx_count);

                if (tx_count_last == m_tx_count)
                {
                    ++m_tx_count_equal;
                }
                else
                {
                    m_tx_count_equal = 0;
                }

                if (m_dsm_under_test->getState() == pm_state_ready ||
                        m_rx_count >= m_dsm_dev_count ||
                        m_tx_count_equal > tx_threshold)
                {
                    m_state = 1;
                    m_total_tx += m_tx_dev_count;
                    m_total_rx += m_rx_count;
                }
                else
                {
                    m_status_mutex.lock();
                    m_status_info = "Check " + util::str(m_dsm_dev_count) +
                            " devices on dsm " + util::str(m_dsm_under_test->getDSID()) +
                            ", " + util::str(m_tx_dev_count) + " of " + util::str(m_dsm_dev_count) + " completed";
                    m_progress = (unsigned char)((m_total_tx + m_tx_dev_count) * 100 / m_total_dev_count);
                    m_status_mutex.unlock();

                    FILE_LOG(logDEBUG1) << "[Task Device Presence] Ping Progress = "
                            << util::str(m_progress) <<
                            ", tx-equal = " << util::str(m_tx_count_equal) <<
                            ", tx-total = " << util::str(m_total_tx) <<
                            ", tx-dev-count = " << util::str(m_tx_dev_count) <<
                            ", tx-count = " << util::str(m_tx_count) <<
                            ", rx-count = " << util::str(m_rx_count);

                    dsa->scheduleLongTask(boost::shared_ptr<dSLongTask> (shared_from_this()), reschedule_time);
                    return;
                }
            }
            break;

        }
    }

    // don't forget to unsubscribe from the listener
    dsa->dS485()->removeListener(this);

    // loop over devices and count successful replies
    size_t total_replies = 0;
    size_t total_removed = 0;
    size_t total_discovered = 0;
    for (std::list<boost::shared_ptr<
            TestDevicePresenceTask::DeviceP> >::iterator it = m_devs->begin();
            it != m_devs->end();
            it++)
    {
        switch ((*it)->getPingState())
        {
            case pd_state_replied:
                total_replies ++;
                break;
            case pd_state_removed:
                Database::getInstance()->updateDevicePresenceStatus(m_task_id,
                        util::str((*it)->getDSMdSID()), (*it)->getId(),
                        (*it)->getPingState());
                total_removed ++;
                break;
            case pd_state_discovered:
                Database::getInstance()->updateDevicePresenceStatus(m_task_id,
                        util::str((*it)->getDSMdSID()), (*it)->getId(),
                        (*it)->getPingState());
                total_discovered ++;
                break;
            default:
                break;
        }
    }

    FILE_LOG(logINFO) << "[Task Device Presence] Finished, "
            "total " << util::str(m_total_dev_count) << " devices checked, "
            "got " << util::str(total_replies) << " replies, " <<
            util::str(total_removed) << " changed to inactive " <<
            util::str(total_discovered) << " changed to active";

    m_status_mutex.lock();
    m_progress = 100;
    if (!m_valid)
    {
        m_task_status = Cancelled;
        m_status_info = "Cancelled";
    }
    else
    {
        m_task_status = Finished;
        m_status_info = "Finished";
    }
    m_status_mutex.unlock();
}

TestDeviceStatisticTask::TestDeviceStatisticTask(dsid_t dsm_dsid,
        dsid_t dev_dsid) :
        dSLongTask(), m_dsm_dsid(dsm_dsid), m_dev_dsid(dev_dsid)
{
    m_name = "Statistik Daten des Gerätes";
    m_description = "Abfrage der Statistikdaten von " + util::str(dev_dsid);
    m_parameters = "dsm=" + util::str(dsm_dsid) + "&dsid=" +
                                                        util::str(dev_dsid);
    m_valid = true;
    m_pausable = true;
    m_state = kStateInit;
}

task_type TestDeviceStatisticTask::getTaskType() const
{
    return TestDeviceStatistic;
}

void TestDeviceStatisticTask::callback(boost::shared_ptr<CBData> busdata)
{
    switch(busdata->getType())
    {
        default:
            break;

        case DeviceStatistics:
        {
            boost::shared_ptr<DeviceStatisticData> data =
                boost::dynamic_pointer_cast<DeviceStatisticData> (busdata);

            dsid_t idm = data->getDSMSourceDSID();
            dsid_t idd = m_device->getDSMdSID();
            if (IsEqualId(idm, idd) &&
                    (data->getDeviceID() == m_device->getId()))
            {
                m_result_value = data->getValue();
                m_result_err = 0;
                m_pstats->setValue(data->getOffset() * 2, m_result_value, 0);
                m_retry = 0;
            }
        }
        break;
    }
}

void TestDeviceStatisticTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        m_status_mutex.lock();
        m_task_status = Cancelled;
        m_status_mutex.unlock();
        return;
    }

    m_status_mutex.lock();
    m_task_status = Running;
    m_status_mutex.unlock();

    if (m_state == kStateInit)
    {
        try
        {
            if (IsNullId(m_dsm_dsid))
            {
                m_dev_list = Storage::getInstance()->getDevices(true);
            }
            else if (IsNullId(m_dev_dsid))
            {
                m_dev_list = Storage::getInstance()->getDevices(m_dsm_dsid, true);
            }
            else
            {
                m_dev_list = boost::shared_ptr<std::list<boost::shared_ptr<Device> > >
                    (new std::list<boost::shared_ptr<Device> >);
                m_dev_list->push_back(
                        Storage::getInstance()->getDevice(m_dev_dsid, m_dsm_dsid));
            }
        }
        catch (SQLite3Exception &sx)
        {
            FILE_LOG(logERROR) << sx.getMessage();
            if (SQLite3::isFatal(sx.getErrorCode()))
            {
                throw sx;
            }
            m_dev_list.reset();
        }
        catch (Exception &ex)
        {
            m_dev_list.reset();
        }
        m_dev_count = 0;
        m_dev_current = m_dev_list->begin();
        m_progress = 0;

        dsa->dS485()->addListener(this);
        this->select(DeviceStatistics, true);

        m_state = kStateSetup;
    }

    if (!m_dev_list || m_dev_list->size() == 0)
    {
        m_status_mutex.lock();
        m_progress = 100;
        m_task_status = Finished;
        m_status_mutex.unlock();
        return;
    }

    if (m_paused)
    {
        m_paused = false;
    }

    while (m_valid && (m_state != kStateFinished))
    {
        if (m_paused)
        {
            m_task_status = Sleeping;
            return;
        }

        if (dsa->getBusState() != DS485_ACTIVE)
        {
            FILE_LOG(logINFO) << "[Task Device Statistic] "
                    "Suspended, dS485 bus down";
            m_task_status = Sleeping;
            dsa->scheduleLongTask(boost::shared_ptr<dSLongTask> (shared_from_this()), 30);
            return;
        }

        switch (m_state)
        {
            case kStateSetup:
            {
                m_device = *m_dev_current;
                m_dev_count ++;
                m_offset = 0;
                m_length = 0;
                m_retry = 0;
                m_pstats.reset();
                m_status_info = "Requesting statistic from device " +
                        util::str(m_device->getDSID());
                FILE_LOG(logINFO) << "[Task Device Statistic] "
                        "Start to query statistic values from device " <<
                        util::str(m_device->getDSID());
                m_state = kStateVersion;
            }
            break;

            case kStateVersion:
            {
                m_retry = 0;
                while (m_retry < 3)
                {
                    try
                    {
                        dsa->dS485()->getStatisticValue(m_device->getDSMdSID(),
                                m_device->getId(), m_offset, &m_result_value);
                        m_pstats = boost::shared_ptr<StatisticDevice>
                            (new StatisticDevice(m_result_value & 0xff));
                        m_pstats->pushValue(m_result_value, 0);
                        m_length = m_result_value >> 9;
                        m_state = kStateRequest;
                        m_offset ++;
                        m_retry = 0;
                        break;
                    }
                    catch (Exception &ex)
                    {
                        m_result_err = ex.getErrorCode();
                        m_retry ++;
                    }
                }
                if (m_retry)
                {
                    FILE_LOG(logINFO) << "[Task Device Statistic] "
                            "Failed to get statistic version from device " <<
                            util::str(m_device->getDSID()) <<
                            " [Error " << util::str(m_result_err) << "]";
                    Database::getInstance()->saveStatisticDevice(m_task_id,
                            m_device->getDSID(), m_device->getDSMdSID(),
                            m_device->getId(), STAT_VERSION, 0, m_result_err);
                    m_state = kStateNext;
                }
            }
            break;

            case kStateRequest:
            {
                m_result_err = ERROR_TIMEOUT;
                m_progress = (unsigned char)
                        (((m_dev_count-1) * 54 + m_offset) * 100 / (m_dev_list->size() * 54));
                try
                {
                    dsa->dS485()->reqStatisticValue(m_device->getDSMdSID(),
                            m_device->getId(), m_offset);
                    m_state = kStateResponse;
                }
                catch (Exception &ex)
                {
                    m_result_err = ex.getErrorCode();
                    m_retry ++;
                }

                if (m_retry >= 3)
                {
                    FILE_LOG(logINFO) << "[Task Device Statistic] "
                            "Failed to get offset " <<
                            util::str(m_offset) << " from device " <<
                            util::str(m_device->getDSID()) <<
                            " [Error " << util::str(m_result_err) << "]";
                    m_pstats->setValue(m_offset * 2, 0, m_result_err);
                    m_offset ++;
                    m_retry = 0;
                }
                dsa->scheduleLongTask(boost::shared_ptr<dSLongTask> (shared_from_this()), 5);
                return;
            }
            break;

            case kStateResponse:
            {
    			m_state = kStateRequest;
            	if (m_result_err)
            	{
            		if ((m_retry++ >= 3) || (m_result_err != ERROR_TIMEOUT))
            		{
            			FILE_LOG(logINFO) << "[Task Device Statistic] "
            					"Failed to get offset " <<
            					util::str(m_offset) << " from device " <<
            					util::str(m_device->getDSID()) <<
            					" [Error " << util::str(m_result_err) << "]";
            			m_pstats->setValue(m_offset * 2, 0, m_result_err);
            			m_offset ++;
            			m_retry = 0;
            		}
        			continue;
            	}
                if (++m_offset >= m_length)
                {
                    m_state = kStateReady;
                }
                m_progress = (unsigned char)
                        (((m_dev_count-1) * 54 + m_offset) * 100 / (m_dev_list->size() * 54));
            }
            break;

            case kStateReady:
            {
                m_pstats->parseValues();

                FILE_LOG(logINFO) << "[Task Device Statistic] "
                        "Query complete got " <<
                        util::str(m_offset) << " values from device " <<
                        util::str(m_device->getDSID());

                Database::getInstance()->saveStatisticDevice(m_task_id,
                        m_device->getDSID(), m_device->getDSMdSID(),
                        m_device->getId(), m_pstats->m_value_map);

                DSID_PREFIX_LOG(logINFO, m_device->getDSID(), "stat_")
                    << "Statistic Query Results:";

                for (std::list<boost::shared_ptr<StatisticValue> >::iterator it =
                        m_pstats->m_value_map->begin();
                        it != m_pstats->m_value_map->end();
                        it++)
                {
                    std::pair<std::string,std::string> desc =
                            Database::getInstance()->getStatisticDescription((*it)->m_field);
                    if ((*it)->m_errorcode)
                    {
                        DSID_PREFIX_LOG(logINFO, m_device->getDSID(), "stat_")
                            << desc.first << " = ERROR " << (*it)->m_errorcode;
                    }
                    else
                    {
                        DSID_PREFIX_LOG(logINFO, m_device->getDSID(), "stat_")
                            << desc.first << " = " << m_pstats->formatValue(*it);
                    }
                }

                m_state = kStateNext;
            }
            break;

            case kStateNext:
            {
                m_state = kStateSetup;
                m_dev_current ++;
                if (m_dev_current == m_dev_list->end())
                {
                    m_state = kStateFinished;
                }
            }
            break;

            default:
                m_state = kStateFinished;
                break;
        }
    }

    dsa->dS485()->removeListener(this);

    m_status_mutex.lock();
    m_progress = 100;
    if (!m_valid)
    {
        m_task_status = Cancelled;
        m_status_info = "Cancelled";
    }
    else
    {
        m_task_status = Finished;
        m_status_info = "Finished";
    }
    m_status_mutex.unlock();
}

TestDeviceStatisticResetTask::TestDeviceStatisticResetTask(dsid_t dsm_dsid,
        dsid_t dev_dsid) :
        dSLongTask(), m_dsm_dsid(dsm_dsid), m_dev_dsid(dev_dsid)
{
    m_name = "Statistik Daten zurücksetzen";
    m_description = "Statistikdaten von " + util::str(dev_dsid) +
                                                               " zurücksetzen";
    m_parameters = "dsm=" + util::str(dsm_dsid) + "&dsid=" +
                                                           util::str(dev_dsid);
    m_valid = true;
    m_pausable = true;
    m_state = kStateInit;
}

task_type TestDeviceStatisticResetTask::getTaskType() const
{
    return TestDeviceStatisticReset;
}

void TestDeviceStatisticResetTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        m_status_mutex.lock();
        m_task_status = Cancelled;
        m_status_mutex.unlock();
        return;
    }

    m_status_mutex.lock();
    m_task_status = Running;
    m_status_mutex.unlock();

    if (m_state == kStateInit)
    {
        try
        {
            if (IsNullId(m_dsm_dsid))
            {
                m_dev_list = Storage::getInstance()->getDevices(true);
            }
            else if (IsNullId(m_dev_dsid))
            {
                m_dev_list = Storage::getInstance()->getDevices(m_dsm_dsid, true);
            }
            else
            {
                m_dev_list = boost::shared_ptr<std::list<boost::shared_ptr<Device> > >
                    (new std::list<boost::shared_ptr<Device> >);
                m_dev_list->push_back(
                        Storage::getInstance()->getDevice(m_dev_dsid, m_dsm_dsid));
            }
        }
        catch (SQLite3Exception &sx)
        {
            FILE_LOG(logERROR) << sx.getMessage();
            if (SQLite3::isFatal(sx.getErrorCode()))
            {
                throw sx;
            }
            m_dev_list.reset();
        }
        catch (Exception &ex)
        {
            m_dev_list.reset();
        }
        m_dev_count = 0;
        m_dev_current = m_dev_list->begin();
        m_progress = 0;
        m_state = kStateSetup;
    }

    if (!m_dev_list || m_dev_list->size() == 0)
    {
        m_status_mutex.lock();
        m_progress = 100;
        m_task_status = Finished;
        m_status_mutex.unlock();
        return;
    }

    if (m_paused)
    {
        m_paused = false;
    }

    while (m_valid && m_state != kStateFinished)
    {
        if (m_paused)
        {
            m_status_mutex.lock();
            m_task_status = Sleeping;
            m_status_mutex.unlock();
            return;
        }

        switch (m_state)
        {
            case kStateSetup:
            {
                m_device = *m_dev_current;
                m_dev_count ++;
                m_retry = 0;
                m_state = kStateRequest;
                m_progress = (unsigned char) ((m_dev_count-1) * 100 / m_dev_list->size());
                m_status_info = "Reset statistic values of device " +
                        util::str(m_device->getDSID());
                FILE_LOG(logINFO) << "[Task Device Statistic Reset] "
                        "Reset statistic values of device " <<
                        util::str(m_device->getDSID());
            }
            break;

            case kStateRequest:
            {
                m_retry = 0;
                m_errorcode = 0;
                while (m_retry < 3)
                {
                    try
                    {
                        dsa->dS485()->reqStatisticValue(m_device->getDSMdSID(),
                                m_device->getId(), 254);
                        m_state = kStateResponse;
                        m_retry = 0;
                        break;
                    }
                    catch (Exception &ex)
                    {
                        m_errorcode = ex.getErrorCode();
                        m_retry ++;
                    }
                }
                if (m_retry)
                {
                    FILE_LOG(logINFO) << "[Task Device Statistic Reset] "
                            "Failed to reset statistic version from device " <<
                            util::str(m_device->getDSID()) <<
                            " [Error " << util::str(m_errorcode) << "]";
                    m_state = kStateNext;
                }
            }
            break;

            case kStateResponse:
            {
                m_state = kStateNext;
                if (m_errorcode != 0)
                {
                    m_state = kStateRequest;
                }
                dsa->scheduleLongTask(boost::shared_ptr<dSLongTask> (shared_from_this()), 3);
                return;
            }
            break;

            case kStateNext:
            {
                m_state = kStateSetup;
                m_dev_current ++;
                if (m_dev_current == m_dev_list->end())
                {
                    m_state = kStateFinished;
                }
            }
            break;

            default:
                m_state = kStateFinished;
                break;
        }
    }

    m_status_mutex.lock();
    m_progress = 100;
    if (!m_valid)
    {
        m_task_status = Cancelled;
        m_status_info = "Cancelled";
    }
    else
    {
        m_task_status = Finished;
        m_status_info = "Finished";
    }
    m_status_mutex.unlock();
}

TestMeterStatisticTask::TestMeterStatisticTask(dsid_t dsm_dsid) :
        dSLongTask(), m_dsm_dsid(dsm_dsid)
{
    m_name = "Statistik Daten des Meters";
    m_description = "Abfrage der Statistikdaten von dSM " +
        util::str(m_dsm_dsid);
    m_parameters = "dsm=" + util::str(m_dsm_dsid);
    m_valid = true;
}

task_type TestMeterStatisticTask::getTaskType() const
{
    return TestMeterStatistic;
}

void TestMeterStatisticTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        m_status_mutex.lock();
        m_task_status = Cancelled;
        m_status_mutex.unlock();
        return;
    }

    m_status_mutex.lock();
    m_task_status = Running;
    m_status_mutex.unlock();

    boost::shared_ptr<std::list<dsid_t> > dsms;
    boost::shared_ptr<std::list<dsid_t> > dsm_list(new std::list<dsid_t>);

    if (IsNullId(m_dsm_dsid))
    {
        dsms = Storage::getInstance()->getDSMList();
    }
    else
    {
        dsms = dsm_list;
        dsms->push_back(m_dsm_dsid);
    }

    size_t total = dsms->size();
    size_t current = 0;
    if (total == 0)
    {
        m_status_mutex.lock();
        m_progress = 100;
        m_task_status = Finished;
        m_status_mutex.unlock();
        return;
    }

    // loop over all dSMeter
    for (std::list<dsid_t>::iterator itm = dsms->begin();
            m_valid && (itm != dsms->end());
            itm++)
    {
        current ++;
        m_status_mutex.lock();
        m_progress = (unsigned char)(current * 100 / total);
        m_status_info = "Abfrage vom Gerät: " + util::str(current) +
            " von " + util::str(total);
        m_status_mutex.unlock();

        FILE_LOG(logINFO) << "[Task Meter Statistic] Started query from dsm " <<
                util::str(*itm);

        uint8_t length;
        uint8_t index;
        uint8_t data[256];
        try
        {
            length = index = 0;
            dsa->dS485()->getStatisticUpstream(*itm, 0, &length, data + index);
            index += length;
            dsa->dS485()->getStatisticUpstream(*itm, 1, &length, data + index);
            index += length;
        }
        catch (Exception& ex)
        {
            FILE_LOG(logERROR) << "[Task Meter Statistic] Query from dsm " <<
                    util::str(*itm) << " failed: " << ex.getMessage();
            continue;
        }
        uint16_t sversion = *(uint16_t *) data;
        boost::shared_ptr<StatisticMeter> pstats(new StatisticMeter(sversion));

        for (int i = 0; i < index; i++)
        {
            pstats->pushValue(data[i], 0);
        }
        pstats->parseValues();

        Database::getInstance()->saveStatisticMeter(m_task_id, *itm, pstats->m_value_map);

        boost::shared_ptr<std::map<statistic_field_t, std::string > > sresults =
            pstats->formatValues();
        for (std::map<statistic_field_t, std::string>::iterator it = sresults->begin();
             it != sresults->end();
             it++)
        {
            std::pair<std::string,std::string> desc =
                    Database::getInstance()->getStatisticDescription(
                                                                (*it).first);
            DSID_PREFIX_LOG(logINFO, *itm, "stat_") << desc.first
                                                    << " = " << (*it).second;
        }
    }

    m_status_mutex.lock();
    if (!m_valid)
    {
        m_task_status = Cancelled;
        m_status_info = "Cancelled";
    }
    else
    {
        m_task_status = Finished;
        m_status_info = "Finished";
    }
    m_status_mutex.unlock();
}

TestMeterStatisticResetTask::TestMeterStatisticResetTask(dsid_t dsm_dsid) :
        dSLongTask(), m_dsm_dsid(dsm_dsid)
{
    m_name = "Statistik Daten des Meters zurücksetzen";
    m_description = "Statistikdaten von dSM " + util::str(m_dsm_dsid) +
                                                               " zurücksetzen";
    m_parameters = "dsm=" + util::str(m_dsm_dsid);
    m_valid = true;
}

task_type TestMeterStatisticResetTask::getTaskType() const
{
    return TestMeterStatisticReset;
}

void TestMeterStatisticResetTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        m_status_mutex.lock();
        m_task_status = Cancelled;
        m_status_mutex.unlock();
        return;
    }

    m_status_mutex.lock();
    m_task_status = Running;
    m_status_mutex.unlock();

    boost::shared_ptr<std::list<dsid_t> > dsms;
    boost::shared_ptr<std::list<dsid_t> > dsm_list(new std::list<dsid_t>);

    if (IsNullId(m_dsm_dsid))
    {
        dsms = Storage::getInstance()->getDSMList();
    }
    else
    {
        dsms = dsm_list;
        dsms->push_back(m_dsm_dsid);
    }

    size_t total = dsms->size();
    size_t current = 0;
    if (total == 0)
    {
        m_status_mutex.lock();
        m_progress = 100;
        m_task_status = Finished;
        m_status_mutex.unlock();
        return;
    }

    // loop over all dSMeter
    for (std::list<dsid_t>::iterator itm = dsms->begin();
            m_valid && (itm != dsms->end());
            itm++)
    {
        current ++;
        m_status_mutex.lock();
        m_progress = (unsigned char)(current * 100 / total);
        m_status_info = "Statistik Reset, Gerät: " + util::str(current) +
            " von " + util::str(total);
        m_status_mutex.unlock();

        try
        {
            dsa->dS485()->resetStatisticUpstream(*itm);
        }
        catch (Exception &ex)
        {
            FILE_LOG(logERROR) << "[Task Meter Statistic Reset] "
                    "Command failed for meter " <<
                    util::str(*itm) << ": " << ex.getMessage();
            continue;
        }
    }

    m_status_mutex.lock();
    if (!m_valid)
    {
        m_task_status = Cancelled;
        m_status_info = "Cancelled";
    }
    else
    {
        m_task_status = Finished;
        m_status_info = "Finished";
    }
    m_status_mutex.unlock();
}

TestSystemMonitorTask::TestSystemMonitorTask(int profile, int duration) :
        dSLongTask(), m_profile(profile), m_duration(duration)
{
    m_name = "System Performance Monitor Test";
    m_description = "Periodische Abfrage von Kommunikations- und Statistikwerten";
    m_parameters = "profile=" + util::str(m_profile) + ";duration=" + util::str(m_duration);
    m_valid = true;
    m_pausable = true;
    m_state = kStateInit;
}

task_type TestSystemMonitorTask::getTaskType() const
{
    return TestSystemMonitor;
}

TestSystemMonitorTask::SystemTestMethod::SystemTestMethod(
        task_type task, dsid_t dsm_dsid, dsid_t device_dsid, time_t dt) :
        m_type(task), m_dsm_dsid(dsm_dsid), m_device_dsid(device_dsid), m_deltatime(dt)
{
	m_lasttime = time(NULL);
}

dSLongTask* TestSystemMonitorTask::SystemTestMethod::getTask()
{
    return m_task.get();
}

std::string TestSystemMonitorTask::SystemTestMethod::getName()
{
    std::string result;
    switch (m_type)
    {
        case TestTransmissionQuality:
            result = "Systemtest Transmission Quality";
            break;
        case TestDevicePresence:
            result = "Systemtest Device Presence";
            break;
        case TestMeterStatistic:
            result = "Systemtest Meter Statistic";
            break;
        case TestDeviceStatistic:
            result = "Systemtest Device Statistic";
            break;
        case TestDeviceStatisticReset:
            result = "Systemtest Device Statistic Reset";
            break;
        case TestMeterStatisticReset:
            result = "Systemtest Meter Statistic Reset";
            break;
        default:
            result = "Invalid Systemtest Task";
            break;
    }
    return result;
}

bool TestSystemMonitorTask::SystemTestMethod::isRunning()
{
    return (m_task && ((m_task->getTaskStatus() == Running) ||
      (m_task->getTaskStatus() == Queued) ||
      (m_task->getTaskStatus() == Sleeping)));
}

bool TestSystemMonitorTask::SystemTestMethod::isToBeScheduled()
{
    if (m_deltatime == 0 || m_deltatime == -1)
    {
        return false;
    }
    return (time(NULL) > m_lasttime + m_deltatime);
}

long long int TestSystemMonitorTask::SystemTestMethod::schedule(long long int parent_id)
{
    if (m_task && m_task.use_count())
    {
        m_task.reset();
    }
    switch (m_type)
    {
        case TestTransmissionQuality:
            m_task = boost::shared_ptr<dSLongTask>
                (new TestTranmissionQualityTask(1, m_dsm_dsid));
            break;
        case TestDevicePresence:
            m_task = boost::shared_ptr<dSLongTask>
                (new TestDevicePresenceTask(m_dsm_dsid, 2000));
            break;
        case TestMeterStatistic:
            m_task = boost::shared_ptr<dSLongTask>
                (new TestMeterStatisticTask(m_dsm_dsid));
            break;
        case TestMeterStatisticReset:
            m_task = boost::shared_ptr<dSLongTask>
                (new TestMeterStatisticResetTask(m_dsm_dsid));
            break;
        case TestDeviceStatistic:
            m_task = boost::shared_ptr<dSLongTask>
                (new TestDeviceStatisticTask(m_dsm_dsid, m_device_dsid));
            break;
        case TestDeviceStatisticReset:
            m_task = boost::shared_ptr<dSLongTask>
                (new TestDeviceStatisticResetTask(m_dsm_dsid, m_device_dsid));
            break;
        default:
            return 0;
    }
    m_task->setParentTaskID(parent_id);
    m_lasttime = time(NULL);
    return dS_Assistant::getInstance()->addLongTask(m_task);
}

void TestSystemMonitorTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        m_status_mutex.lock();
        m_task_status = Cancelled;
        m_status_mutex.unlock();
        return;
    }

    m_status_mutex.lock();
    m_task_status = Running;
    m_status_mutex.unlock();

    // initial setup
    if (kStateInit == m_state)
    {
        m_startup_methods = boost::shared_ptr<std::list<boost::shared_ptr<SystemTestMethod> > >
            (new std::list<boost::shared_ptr<SystemTestMethod> >);
        m_test_methods = boost::shared_ptr<std::list<boost::shared_ptr<SystemTestMethod> > >
            (new std::list<boost::shared_ptr<SystemTestMethod> >);
        m_cleanup_methods = boost::shared_ptr<std::list<boost::shared_ptr<SystemTestMethod> > >
            (new std::list<boost::shared_ptr<SystemTestMethod> >);

        dsid_t NullId;
        SetNullId(NullId);

        switch (m_profile)
        {
            default:
            case 1:
            {
                m_startup_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestMeterStatisticReset, NullId, NullId, 0)));
                m_startup_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestDeviceStatisticReset, NullId, NullId, 0)));

                m_test_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestTransmissionQuality, NullId, NullId, 10 * 60)));
                m_test_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestDevicePresence, NullId, NullId, 20 * 60)));

                m_cleanup_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestMeterStatistic, NullId, NullId, 0)));
                m_cleanup_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestDeviceStatistic, NullId, NullId, 0)));
            }
            break;

            case 2:
            {
                m_startup_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestMeterStatisticReset, NullId, NullId, 0)));
                m_startup_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestDeviceStatisticReset, NullId, NullId, 0)));

                m_test_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestTransmissionQuality, NullId, NullId, 1 * 60 * 60)));
                m_test_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestDevicePresence, NullId, NullId, 6 * 60 * 60)));
                m_test_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestMeterStatistic, NullId, NullId, 1 * 60 * 60)));
                m_test_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestDeviceStatistic, NullId, NullId, 24 * 60 * 60)));
            }
            break;

            case 3:
            {
                m_startup_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestMeterStatisticReset, NullId, NullId, 0)));
                m_startup_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestDeviceStatisticReset, NullId, NullId, 0)));

                m_test_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestTransmissionQuality, NullId, NullId, 15 * 60)));
                m_test_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestDevicePresence, NullId, NullId, 12 * 60 * 60)));
                m_test_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestMeterStatistic, NullId, NullId, 20 * 60)));
                m_test_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestDeviceStatistic, NullId, NullId, 6 * 60 * 60)));
            }
            break;

            case 4:
            {
                m_startup_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestMeterStatisticReset, NullId, NullId, 0)));

                m_test_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestTransmissionQuality, NullId, NullId, 15 * 60)));
                m_test_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestDevicePresence, NullId, NullId, 20 * 60)));
                m_test_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestMeterStatistic, NullId, NullId, 15 * 60)));
                m_test_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestDeviceStatistic, NullId, NullId, 45 * 60)));

                m_cleanup_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestMeterStatistic, NullId, NullId, 0)));
                m_cleanup_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestDeviceStatistic, NullId, NullId, 0)));
            }
            break;

        }

        m_state = kStateStartup;

        FILE_LOG(logINFO) << "[Profile Task] starting profile " <<
        		util::str(m_profile) << " for " <<
        		util::str(m_duration) << " seconds";
    }

    // resuming from paused state?
    if (m_paused)
    {
        m_paused = false;
    }

    // times for next execution of this control handler
    time_t next_schedule = 60;
    time_t next_task_schedule;

    // worker loop
    while (m_valid && (m_state != kStateFinished))
    {
        if (m_paused)
        {
            m_status_mutex.lock();
            m_task_status = Sleeping;
            m_status_mutex.unlock();
            return;
        }

        switch (m_state)
        {
            case kStateStartup:
            {
                // schedule tasks to run once at the begin of the tests
                for (std::list<boost::shared_ptr<SystemTestMethod> >::iterator
                        it = m_startup_methods->begin();
                        it != m_startup_methods->end();
                        it++)
                {
                    boost::shared_ptr<SystemTestMethod> tm(*it);
                    FILE_LOG(logDEBUG3) << "[Profile Task] schedule startup task: " << tm->getName();
                    tm->schedule(getTaskID());
                }
                m_state = kStateStartupWait;
                m_task_status = Sleeping;
                dsa->scheduleLongTask(boost::shared_ptr<dSLongTask> (shared_from_this()), 30);
                return;
            }
            break;

            case kStateStartupWait:
            {
                int active_tasks = 0;
                for (std::list<boost::shared_ptr<SystemTestMethod> >::iterator
                        it = m_startup_methods->begin();
                        it != m_startup_methods->end();
                        it++)
                {
                    boost::shared_ptr<SystemTestMethod> tm(*it);
                    if (tm->isRunning())
                    {
                        active_tasks ++;
                    }
                }
                if (active_tasks == 0)
                {
                    m_state = kStateReady;
                }
                else
                {
                    m_task_status = Sleeping;
                    dsa->scheduleLongTask(boost::shared_ptr<dSLongTask> (shared_from_this()), 30);
                    return;
                }
            }
            break;

            case kStateReady:
            {
                next_schedule = 60;
                time_t now = time(NULL);
                for (std::list<boost::shared_ptr<SystemTestMethod> >::iterator
                        it = m_test_methods->begin();
                        it != m_test_methods->end();
                        it++)
                {
                    boost::shared_ptr<SystemTestMethod> tm(*it);

                    next_task_schedule = 60;
                    if (tm->isRunning())
                    {
                        FILE_LOG(logDEBUG3)
                            << "[Profile Task] still running: "
                            << tm->getName();
                        next_task_schedule = 10;
                    }
                    else if (now >= tm->m_lasttime + tm->m_deltatime)
                    {
                        FILE_LOG(logDEBUG3)
                            << "[Profile Task] schedule subtask: "
                            << tm->getName();
                        tm->schedule(getTaskID());
                    }
                    else
                    {
                        next_task_schedule = tm->m_lasttime + tm->m_deltatime - now;
                        FILE_LOG(logDEBUG3)
                        	<< "[Profile Task] subtask " << tm->getName()
                        	<< " to be scheduled in "
                        	<< util::str(next_task_schedule);
                    }
                    if (next_schedule > next_task_schedule)
                    {
                        next_schedule = next_task_schedule;
                    }
                }
                m_state = kStateWait;
            }
            break;

            case kStateWait:
            {
                m_state = kStateReady;
                time_t now = time(NULL);
                if (now - m_started >= m_duration)
                {
                    FILE_LOG(logDEBUG3)
                        << "[Profile Task] main test time over";;
                    m_state = kStateCleanup;
                }
                else if (next_schedule > 0)
                {
                    m_task_status = Sleeping;
                    dsa->scheduleLongTask(boost::shared_ptr<dSLongTask> (shared_from_this()), next_schedule);
                    FILE_LOG(logDEBUG3)
                        << "[Profile Task] sleep for "
                        << util::str(next_schedule);
                    return;
                }
            }
            break;

            case kStateCleanup:
            {
                int active_tasks = 0;
                for (std::list<boost::shared_ptr<SystemTestMethod> >::iterator
                        it = m_test_methods->begin();
                        it != m_test_methods->end();
                        it++)
                {
                    boost::shared_ptr<SystemTestMethod> tm(*it);
                    if (tm->isRunning())
                    {
                        active_tasks ++;
                    }
                }
                if (active_tasks == 0)
                {
                    // schedule tasks to run once at the end of the tests
                    for (std::list<boost::shared_ptr<SystemTestMethod> >::iterator
                            it = m_cleanup_methods->begin();
                            it != m_cleanup_methods->end();
                            it++)
                    {
                        boost::shared_ptr<SystemTestMethod> tm(*it);
                        FILE_LOG(logDEBUG3) << "[Profile Task] schedule cleanup task: " << tm->getName();
                        tm->schedule(getTaskID());
                    }
                    m_state = kStateCleanupWait;
                }
                m_task_status = Sleeping;
                dsa->scheduleLongTask(boost::shared_ptr<dSLongTask> (shared_from_this()), 10);
                return;
            }
            break;

            case kStateCleanupWait:
            {
                int active_tasks = 0;
                for (std::list<boost::shared_ptr<SystemTestMethod> >::iterator
                        it = m_cleanup_methods->begin();
                        it != m_cleanup_methods->end();
                        it++)
                {
                    boost::shared_ptr<SystemTestMethod> tm(*it);
                    if (tm->isRunning())
                    {
                        active_tasks ++;
                    }
                }
                if (active_tasks == 0)
                {
                    m_state = kStateFinished;
                }
                else
                {
                    m_task_status = Sleeping;
                    dsa->scheduleLongTask(boost::shared_ptr<dSLongTask> (shared_from_this()), 10);
                    return;
                }
            }
            break;

            default:
                m_state = kStateFinished;
                break;
        }
    }

    FILE_LOG(logINFO) << "[Profile Task] ready";

    m_status_mutex.lock();
    if (!m_valid)
    {
        m_task_status = Cancelled;
        m_status_info = "Cancelled";
    }
    else
    {
        m_task_status = Finished;
        m_status_info = "Finished";
    }
    m_status_mutex.unlock();
}


BasicMonitorTask::BasicMonitorTask() : dSTask()
{
    m_name = "Basic traffic monitor";
    m_description = "Permanently monitor and log ds traffic";
    m_valid = true;
    m_state = kStateInit;

    // TODO: implement storage enabler for system test tasks
    m_db_storage = false;
}

void BasicMonitorTask::callback(boost::shared_ptr<CBData> busdata)
{
    boost::shared_ptr<CBData> data = busdata;

    switch(busdata->getType())
    {
        default:
            break;

        case CommUpstreamShort:
        {
            boost::shared_ptr<CommUpstreamShortData> data =
                boost::dynamic_pointer_cast<CommUpstreamShortData> (busdata);

            std::string decoded = dsutil::rkDecodeShort(
                    data->getDSMSourceDSID(),
                    data->getDeviceID(),
                    data->isResend(),
                    data->isSensor(),
                    data->getKeyType(),
                    data->getClickType(),
                    data->getFlags(),
                    data->getCrosstalkFlags());

            DSID_LOG(logINFO, data->getDSMSourceDSID()) <<
                "UP-S [Q:" << std::setprecision (2) << (data->getQuality() * 100.0/62 + .5) <<
                ", P:" << ((data->getFlags() & 1) > 0 ? 'N' : 'P') <<
                ", C:" << ((data->getCrosstalkFlags() & 1) > 0 ? '1' : '0') <<
                "] " << decoded;
        }
        break;

        case CommUpstreamLong:
        {
            boost::shared_ptr<CommUpstreamLongData> data =
                boost::dynamic_pointer_cast<CommUpstreamLongData> (busdata);

            std::string decoded = dsutil::rkDecodeLong(
                    data->getDSMSourceDSID(),
                    data->getDeviceID(),
                    data->getMainQualifier(),
                    data->getSubQualifier(),
                    data->getData(),
                    data->getFlags(),
                    data->getCrosstalkFlags());

            DSID_LOG(logINFO, util::str(data->getDSMSourceDSID())) <<
                "UP-L [Q:" << std::setprecision (2) << (data->getQuality() * 100.0/62 + .5) <<
                ", P:" << ((data->getFlags() & 1) > 0 ? 'N' : 'P') <<
                ", C:" << ((data->getCrosstalkFlags() & 1) > 0 ? '1' : '0') <<
                "] " << decoded;
        }
        break;

        case CommDownstreamShort:
        {
            boost::shared_ptr<CommDownstreamShortData> data =
                boost::dynamic_pointer_cast<CommDownstreamShortData> (busdata);

            DSID_LOG(logINFO, util::str(data->getDSMSourceDSID())) <<
                        dsutil::hkDecodeShort(data->getCommand());
        }
        break;

        case CommDownstreamLong0:
        {
            boost::shared_ptr<CommDownstreamLong0Data> data(
                boost::dynamic_pointer_cast<CommDownstreamLong0Data> (busdata));

            DSID_LOG(logINFO, util::str(data->getDSMSourceDSID())) <<
                        dsutil::hkDecodeLong0(data->getCommand());
        }
        break;

        case CommDownstreamLong8:
        {
            boost::shared_ptr<CommDownstreamLong8Data> data(
                    boost::dynamic_pointer_cast<CommDownstreamLong8Data> (busdata));

            DSID_LOG(logINFO, util::str(data->getDSMSourceDSID())) <<
                        dsutil::hkDecodeLong8(data->getCommand(), data->getData());
        }
        break;

        case CommDownstreamLong16:
        {
            boost::shared_ptr<CommDownstreamLong16Data> data(
                boost::dynamic_pointer_cast<CommDownstreamLong16Data>(busdata));

            DSID_LOG(logINFO, util::str(data->getDSMSourceDSID())) <<
                        dsutil::hkDecodeLong16(data->getCommand(), data->getData());
        }
        break;

    }

    if (m_db_storage)
    {
        m_mutex.lock();
        m_queue.push(data);
        m_mutex.unlock();
    }
}

void BasicMonitorTask::process()
{
    boost::shared_ptr<CBData> busdata;
    boost::shared_ptr<std::list<boost::shared_ptr<CBData> > > cbdatalist(
            new(std::list<boost::shared_ptr<CBData> >));

    m_mutex.lock();
    while (!m_queue.empty())
    {
        busdata = m_queue.front();
        m_queue.pop();
        cbdatalist->push_back(busdata);
    }
    m_mutex.unlock();

    if (cbdatalist->size())
    {
        Database::getInstance()->saveUpstreamMetric(cbdatalist);
    }
}

void BasicMonitorTask::run(dS_Assistant *dsa)
{
    // initial setup
    if (kStateInit == m_state)
    {
        // add task to receive callbacks
        dsa->dS485()->addListener(this);

        // select the callbacks you want
        this->select(CommUpstreamShort, true);
        this->select(CommUpstreamLong, true);
        this->select(CommDownstreamShort, true);
        this->select(CommDownstreamLong0, true);
        this->select(CommDownstreamLong8, true);
        this->select(CommDownstreamLong16, true);

        m_state = kStateSetup;
    }

    while (m_valid)
    {
        switch (m_state)
        {
            case kStateSetup:
            {
                process();
                m_state = kStateReady;
            }
            break;

            case kStateReady:
            default:
            {
                m_state = kStateSetup;
                dsa->scheduleTask(boost::shared_ptr<dSTask> (shared_from_this()), 5);
                return;
            }
            break;
        }
    }

    // final cleanup
    dsa->dS485()->removeListener(this);
}


DatabaseCleanupTask::DatabaseCleanupTask(size_t max_db_filesize) :
        dSTask(), m_max_filesize(max_db_filesize)
{
    m_name = "Periodical database cleanup";
    m_description = "Cleanup and database maintenance";
    m_valid = true;
    m_state = kStateInit;
}

void DatabaseCleanupTask::run(dS_Assistant *dsa)
{
    // initial setup
    if (kStateInit == m_state)
    {
        FILE_LOG(logINFO) << "[DatabaseCleanup] " <<
                "Starting cleanup function, current db size: " <<
                util::str(Database::getInstance()->getFileSize());
        m_iteration = 0;
        m_state = kStateSetup;
    }

    while (m_valid)
    {
        switch (m_state)
        {
            case kStateSetup:
            {
                m_iteration ++;
                Database* db = Database::getInstance();
                if (db->getFileSize() > m_max_filesize)
                {
                    switch(m_iteration)
                    {
                        case 1:
                            db->removeUpstreamMetricByTime(DB_CLEANUP_DAYS(21));
                            break;
                        case 2:
                            db->removeUpstreamMetricByTime(DB_CLEANUP_DAYS(7));
                            break;
                        case 3:
                            db->removeUpstreamMetricByTime(DB_CLEANUP_DAYS(3));
                            break;
                        default:
                            m_state = kStateReady;
                    }
                    db->executeVacuum();
                }
                if (db->getFileSize() < m_max_filesize)
                {
                    m_state = kStateReady;
                }
            }
            break;

            case kStateReady:
            default:
            {
                // calculate next execution time, 7 days ahead at 4am
                time_t now = time(NULL);
                struct tm tmp;
                localtime_r(&now, &tmp);
                tmp.tm_mday += 7;
                tmp.tm_hour = 4;
                tmp.tm_min = 0;

                size_t dbsize = Database::getInstance()->getFileSize();
                if (dbsize > m_max_filesize)
                {
                    FILE_LOG(logWARNING) << "[DatabaseCleanup] " <<
                            "file size still too large: " <<
                            util::str(dbsize);
                    localtime_r(&now, &tmp);
                    tmp.tm_mday ++;
                    tmp.tm_hour = 4;
                    tmp.tm_min = 0;
                }
                time_t scheduledTime = mktime(&tmp);

                FILE_LOG(logINFO) << "[DatabaseCleanup] " <<
                        "schedule next maintenance in " <<
                        util::str((scheduledTime - now) / 60) << " minutes.";

                boost::shared_ptr<DatabaseCleanupTask>
                                task(new DatabaseCleanupTask(m_max_filesize));
                if (scheduledTime > now)
                {
                    dsa->scheduleTask(task, scheduledTime - now);
                }
                else
                {
                    dsa->scheduleTask(task, 12 * 60 * 60);
                }
                return;
            }
            break;
        }
    }

    // final cleanup

}


} // namespace
