/*
 Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

 Author: Michael Tross <mt@aizo.ag>

 This file is part of dS Assistant.

 dS-Assistant is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 dS-Assistant is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

 NOTE: all APIs, especially the JSON API, offered by this application
 are subject to change without further notice.
 */

#ifdef HAVE_CONFIG_H
#include "autoconfig.h"
#endif

#include <sstream>
#include <fstream>
#include <iomanip>
#include <arpa/inet.h>
#include <boost/shared_ptr.hpp>
#include <digitalSTROM/ds.h>
#include "exceptions.h"
#include "hwdesc.h"
#include "storage.h"
#include "traffic-decode.h"
#include "util.h"
#include "log.h"

namespace dSA
{

static const char* keyNumber2String[16] =
{
        "Erste Taste",
        "Zweite Taste",
        "Dritte Taste",
        "Vierte Taste",
        "Erster Lokaltaster",
        "Zweiter Lokaltaster",
        "Dritter Lokaltaster",
        "Vierter Lokaltaster",
        "Taste-8",
        "Taste-9",
        "Taste-10",
        "UDI Interrupt",
        "Taste-12",
        "Testsender",
        "Taste-14",
        "Farbauswahl"
};

static const char* clickTypeToString[16] =
{
        "CT_1T Tipp-1",
        "CT_2T Tipp-2",
        "CT_3T Tipp-3",
        "CT_4T Tipp-4",
        "CT_HS Halten-Start",
        "CT_HR Halten-Repeat",
        "CT_HE Halten-End",
        "CT_1C Kurzer Klick",
        "CT_2C Doppelklick",
        "CT_3C Dreifachklick",
        "CT_1P Kurz-Lang-Klick",
        "CT_L0 Lokaltaster aus",
        "CT_L1 Lokaltaster an",
        "CT_2P Kurz-Kurz-Lang-Klick",
        "CT_LS Lokaltaster stop",
        "Type-15"
};

static const char* groupToString[16] =
{
    "Broadcast", "Licht", "Schatten", "Klima", "Audio", "Video", "Sicherheit",
    "Zugang", "Joker", "Weiße Ware", "Gruppe 10", "Gruppe 11", "Gruppe 12",
    "Gruppe 13", "Gruppe 14", "Gruppe 15"
};

std::string dsutil::hkDecodeShort(uint8_t msg)
{
    std::string buf;
    struct groupSceneS
    {
        uint8_t group;
        uint8_t scene;
    };
    static struct groupSceneS downstream_short_table[42] =
    {
    { 1, 0 },
    { 1, 5 },
    { 1, 17 },
    { 1, 18 },
    { 1, 19 },
    { 1, 1 },
    { 1, 6 },
    { 1, 20 },
    { 1, 21 },
    { 1, 22 },
    { 1, 2 },
    { 1, 7 },
    { 1, 23 },
    { 1, 24 },
    { 1, 25 },
    { 1, 11 },
    { 1, 12 },
    { 1, 255 },
    { 2, 0 },
    { 2, 5 },
    { 2, 1 },
    { 2, 6 },
    { 2, 15 },
    { 2, 255 },
    { 3, 255 },
    { 4, 0 },
    { 4, 5 },
    { 4, 1 },
    { 4, 6 },
    { 4, 255 },
    { 5, 0 },
    { 5, 5 },
    { 5, 1 },
    { 5, 6 },
    { 5, 255 },
    { 6, 0 },
    { 6, 5 },
    { 6, 255 },
    { 7, 0 },
    { 7, 5 },
    { 7, 255 }
    };

    buf = "DN-S ";

    if (msg <= 162)
    {
        uint8_t zone = msg / 41 + 1;
        uint8_t index = msg % 41;
        buf += "Select_A: VSK " + util::str(zone) +
                " Group " + util::str(downstream_short_table[index].group) +
                "(" +
                std::string(groupToString[downstream_short_table[index].group])+
                ")";
        if (downstream_short_table[index].scene != 255)
        {
            buf += " and CallScene " +
                   util::str(downstream_short_table[index].scene);
        }
    }
    else if (msg == 163)
    {
        buf += "General Ack";
    }
    else
    {
        buf += "Select_A: CallScene " + util::str(msg - 164) +
               " (or " + util::str(msg - 164 + 64) + ")";
    }
    return buf;
}

std::string dsutil::hkDecodeLong0(uint8_t msg)
{
    std::string buf;

    buf = "DN-L ";

    switch (msg)
    {
        case 0x00:
            buf += "ProgON_B";
            break;
        case 0x01:
            buf += "ProgOFF_B";
            break;
        case 0x02:
            buf += "Ping_B";
            break;
        case 0x05:
            buf += "Blink_B";
            break;
        case 0x80:
            buf += "UndoScene_A";
            break;
        case 0x81:
            buf += "ProgON_A";
            break;
        case 0x82:
            buf += "ProgOFF_A";
            break;
        case 0x83:
            buf += "Blink_A";
            break;
        default:
            buf += "Unknown command (no data): " + util::str(msg);
            break;
    }
    return buf;
}

std::string dsutil::hkDecodeLong8(uint8_t msg, uint8_t data)
{
    std::string buf;

    buf = "DN-L ";

    switch (msg)
    {
        case 0x00:
            buf += "AddGrp_A: Group " + util::str(data);
            break;
        case 0x01:
            buf += "RemGrp_A: Group " + util::str(data);
            break;
        case 0x02:
            buf += "CallScene_A: Scene " + util::str(data);
            break;
        case 0x03:
            buf += "SaveScene_A: Scene " + util::str(data);
            break;
        case 0x04:
            buf += "SetOutval_A: Value " + util::str(data);
            break;
        case 0x05:
            buf += "UndoSceneNr_A: Scene " + util::str(data);
            break;
        case 0x06:
            buf += "ForceCallScene_A: Scene " + util::str(data);
            break;
        case 0x07:
            buf += "SetLocalPrio_A: Scene " + util::str(data);
            break;
        case 0x08:
            buf += "CallSceneMin_A: Scene " + util::str(data);
            break;
        case 0x1a:
            buf += "TransmissionQuality: Id " + util::str(data >> 6) +
                ", Quality " + util::str(data & 0x3f);
            break;
        case 0x1b:
            buf += "SetCircuitId: " + util::str(data);
            break;
        case 0x1c:
            buf += "SetDsmDsid0: " + util::str(data);
            break;
        case 0x1d:
            buf += "SetDsmDsid1: " + util::str(data);
            break;
        case 0x1e:
            buf += "SetDsmDsid2: " + util::str(data);
            break;
        case 0x1f:
            buf += "SetDsmDsid3: " + util::str(data);
            break;
        case 0x20:
            buf += "CallScene_B: Scene " + util::str(data);
            break;
        case 0x21:
            buf += "ReadPara16_B: Parameter " + util::str(data);
            break;
        case 0x22:
            buf += "Jumble_B: Startaddress " +
                   util::str(((data >> 4) + 1) * 16) +
                   ", Range " + util::str(((data & 0xF) + 1) * 16);
            break;
        case 0x23:
            buf += "Test_B: Testdata " + util::str(data);
            break;
        case 0x24:
            buf += "UDI_RW_B: UDI-Data: " + util::str(data);
            break;
        case 0x25:
            buf += "UDI_W_B: UDI-Data: " + util::str(data);
            break;
        case 0x26:
            buf += "UDI_Last_RW_B: UDI-Data: " + util::str(data);
            break;
        case 0x27:
            buf += "UDI_Last_W_B: UDI-Data: " + util::str(data);
            break;
        case 0x28:
            buf += "BankSelect_B: Bank " + util::str(data);
            break;
        case 0x29:
            buf += "GetSensorValue_B: SensorId " + util::str(data);
            break;
        case 0x2B:
            buf += "GetSensorType_B: SensorId " + util::str(data);
            break;
        case 0x33:
            buf += "GetRamstats_B: Address " + util::str(data);
            break;
        default:
            buf += "Unknown command (8 bit data): " + util::str(msg) +
                   ", data: " + util::str(data);
            break;
    }
    return buf;
}

static std::string parseAddressing(uint16_t data)
{
    uint16_t address = data >> 6;
    uint16_t group_subaddress = data & 0x3F;
    if (address < 16)
    {
        return "VSK " + util::str(address) +
               ", Group " + util::str(group_subaddress);
    }
    else
    {
        return "Device " + util::str(address) + "." +
               util::str(group_subaddress);
    }
}

std::string dsutil::hkDecodeLong16(uint8_t msg, uint16_t data)
{
    std::string buf;

    buf = "DN-L ";
    switch (msg)
    {
        case 0:
            buf += "Select_A: " + parseAddressing(data);
            break;
        case 1:
            buf += "Select_B: " + parseAddressing(data);
            break;
        case 2:
            buf += "Select_DSIDH_B: " + util::hex(data, 4); break;
            break;
        case 3:
            buf += "Select_DSIDL_B: " + util::hex(data, 4); break;
            break;
        case 4:
            buf += "SetPara_B: Address " + util::hex(data >> 8, 2) +
                ", Value " + util::hex(data & 0xff, 2); break;
            break;
        case 5:
            buf += "SensorPush: Sensortype " + util::hex((data >> 10) & 0x3f, 2) +
                ", Value " + util::hex(data & 0x3ff, 4); break;
            break;
        case 6:
            buf += "SetAdr_B: " + util::str(data >> 6) + "." +
                util::str(data & 0x3F);
            break;
        case 7:
            {
                uint16_t scene = data >> 10;
                uint16_t group = data & 0x3F;
                if (group == 0)
                {
                    scene += 64;
                }
                buf += "Select_Gr_A:" + parseAddressing(data & 0x3FF) +
                       " and Call Scene: " + util::str(scene);
            }
            break;
        default:
            buf += "Unknown command (16 bit data): " + util::str(msg) +
                   ", data: " + util::str(data);
            break;
    }
    return buf;
}

std::string dsutil::rkDecodeShort(dsid_t src_dsm_dsid, uint16_t device_id,
        uint8_t is_resend, uint8_t is_sensor,
        uint8_t key_type, uint8_t click_type,
        uint8_t flags, uint8_t cflags)
{
    std::string buf;
    uint8_t data = ((key_type & 15) << 4) | (click_type & 15);
    boost::shared_ptr<const Device> dsd;

    try
    {
        dsd = Storage::getInstance()->getDevice(device_id, src_dsm_dsid);
    }
    catch(Exception &ex)
    {
    }

    buf = "Device " + util::str(device_id) + ": ";

    if (cflags > 0)
    {
        buf += "Data:" + util::hex((uint32_t)(key_type << 8 | click_type |
                is_resend << 16 | is_sensor << 24));
        return buf;
    }

    if (data == 0xff)
    {
        buf = "HELLO from " + util::str(device_id);
    }
    else if(click_type == 13)
    {
        buf = "Testsender Id " + util::str(key_type);
    }
    else if (NULL == dsd)
    {
        buf = "Unknown Device at address " + util::str(device_id) +
              ", Button " + util::str(key_type) + ", Click " +
              util::str(click_type);
    }

    /* legacy R1.0 special devices */
    else if (dsd->getFunctionId() == 0x7050)
    {
        buf += "dSID " + util::str(dsd->getDSID());
        buf += ", Function(leave home)";
        buf += ", Button " + util::str(key_type) +
               " (" + keyNumber2String[key_type] + ")";
        buf += ", Click " + util::str(click_type) +
               " (" + clickTypeToString[click_type] + ")";
        if (is_resend)
        {
            buf += ", Resend";
        }
    }
    else if (dsd->getFunctionId() == 0x7010)
    {
        buf += "dSID " + util::str(dsd->getDSID());
        buf += ", Function(door bell)";
        buf += ", Button " + util::str(key_type) +
               " (" + keyNumber2String[key_type] + ")";
        buf += ", Click " + util::str(click_type) +
               " (" + clickTypeToString[click_type] + ")";
        if (is_resend)
        {
            buf += ", Resend";
        }
    }
    else if (dsd->getFunctionId() == 0x6001)
    {
        buf += "dSID " + util::str(dsd->getDSID());
        buf += ", Function(panic)";
        buf += ", Button " + util::str(key_type) +
               " (" + keyNumber2String[key_type] + ")";
        buf += ", Click " + util::str(click_type) +
               " (" + clickTypeToString[click_type] + ")";
        if (is_resend)
        {
            buf += ", Resend";
        }
    }

    /* Class 0 */
    else if (FidGetClass(dsd->getFunctionId()) == 0)
    {
        /* Subclass 0: Color Scene Device */
        if (FidGetSubclass(dsd->getFunctionId()) == 0)
        {
            buf += "dSID " + util::str(dsd->getDSID());
            buf += ", Function(color scene device)";
            if (is_sensor)
            {
                buf += ", Sensorindex " + util::str(key_type);
                buf += ", Sensorvalue " + util::str(click_type);
                if (is_resend)
                {
                    buf += ", Resend";
                }
            }
            else
            {
                uint8_t color = (data >> 7) & 0x7;
                uint8_t scene = (data << 1) | is_resend;
                buf += ", Color " + util::str(color) +
                        " (" + groupToString[color] + ")";
                buf += ", Scene " + util::str(scene);
            }
        }
    }

    /* generic device model with identical subclass for each group */
    else if ((FidGetSubclass(dsd->getFunctionId()) == 0) ||
             (FidGetSubclass(dsd->getFunctionId()) == 4))
    {
        uint16_t fid = dsd->getFunctionId();
        std::string fidtuple(util::hex(FidGetClass(fid),1) + "/" +
                util::hex(FidGetSubclass(fid),2) + "/" +
                util::hex(FidGetFunctionModule(fid),2));
        buf += "dSID " + util::str(dsd->getDSID());
        buf += ", Function(" + fidtuple + ")";
        if (is_sensor)
        {
            buf += ", Sensorindex " + util::str(key_type);
            buf += ", Sensorvalue " + util::str(click_type);
        }
        else
        {
            buf += ", Button " + util::str(key_type) +
               " (" + keyNumber2String[key_type] + ")";
            buf += ", Click " + util::str(click_type) +
               " (" + clickTypeToString[click_type] + ")";
        }
        if (is_resend)
        {
            buf += ", Resend";
        }
    }
    else if (FidGetSubclass(dsd->getFunctionId()) == 5)
    {
        uint16_t fid = dsd->getFunctionId();
        std::string fidtuple(util::hex(FidGetClass(fid),1) + "/" +
                util::hex(FidGetSubclass(fid),2) + "/" +
                util::hex(FidGetFunctionModule(fid),2));
        buf += "dSID " + util::str(dsd->getDSID());
        buf += ", Function(scene device:" + fidtuple + ")";
        if (is_sensor)
        {
            buf += ", Sensorindex " + util::str(key_type);
            buf += ", Sensorvalue " + util::str(click_type);
            if (is_resend)
            {
                buf += ", Resend";
            }
        }
        else
        {
            uint8_t cmdtype = is_resend | ((data & 0x1) << 1);
            switch (cmdtype)
            {
                case 0x00: buf += ", Call Scene "; break;
                case 0x01: buf += ", Undo Scene "; break;
                case 0x02: buf += ", Force-Call Scene "; break;
                case 0x03: buf += ", Reserved Command "; break;
            }
            buf += util::str((data >> 1) & 0x7f);
        }
    }

    else
    {
        buf = "Unknown packet type from dSID:" + util::str(dsd->getDSID()) +
              "(address: " + util::str(device_id) + ", function: " +
              util::hex(dsd->getFunctionId()) +  ", data: " +
              util::hex(data) + ")";
    }

    return buf;
}

std::string dsutil::rkDecodeLong(dsid_t src_dsm_dsid, uint16_t device_id,
        uint8_t mainqualifier, uint8_t subqualifier, uint32_t data,
        uint8_t flags, uint8_t cflags)
{
    std::string buf;
    std::string devid;
    boost::shared_ptr<const Device> dsd;

    try
    {
        dsd = Storage::getInstance()->getDevice(device_id, src_dsm_dsid);
    }
    catch (Exception &ex)
    {
    }

    buf = " ";

    if (cflags)
    {
        if (mainqualifier == 15)
        {
            buf += std::string("Data " + util::hex(data));
        }
        else
        {
            buf += std::string("Device " + util::str(device_id) + ", Data " +
                    util::hex((uint32_t)
                            (mainqualifier << 24 | subqualifier << 16 | data)));
        }
        return buf;
    }

    if (NULL == dsd)
    {
        devid = "Unknown Device [Data:" + util::hex(data) + "]";
    }
    else
    {
        devid = util::str(dsd->getDSID());
    }

    if (mainqualifier == 15)
    {
        buf += "Response DSID " + util::hex(data);
    }
    else if (mainqualifier == 14)
    {
        buf += "Device: " + util::str(device_id);
        buf += " (dSID:" + devid + ")";
        buf += ", Statistic Offset " + util::hex(subqualifier, 2) +
               ", Value " + util::hex(data, 4);
    }
    else if (mainqualifier == 12)
    {
        buf += "Give Up DSID " + util::hex(data);
    }
    else if (mainqualifier == 2)
    {
        buf += "Device: " + util::str(device_id);
        buf += " (dSID:" + devid + ") ";
        switch (subqualifier)
        {
            case 0:
                {
                    uint8_t subsubq = (data >> 8) & 0xff;
                    switch (subsubq)
                    {
                        case 0:
                            {
                                buf += "Test_B Response: BER " +
                                       util::hex(data & 0xff, 2);
                            }
                            break;
                        case 1:
                            {
                                buf += "UDI Response: Data " +
                                       util::hex(data & 0xff, 2);
                            }
                            break;
                        default:
                            {
                                buf += "Subqualifier " +
                                       util::hex(subqualifier, 2) +
                                       ", Unknown data type " + util::hex(data);
                            }
                            break;
                    }
                }
                break;
            case 1:
                {
                    // sensortype
                    uint8_t sflags = (data >> 8) & 0xff;
                    uint8_t stype = (data & 0xff);
                    buf += "Sensortype: Index " + util::str(sflags & 0x0f);
                    buf += ", Flags " + util::hex((sflags >> 4) & 0x0f) + " (";
                    if ((sflags & 0x80) == 0)
                    {
                        buf += "Last-Index ";
                    }
                    if ((sflags & 0x40) == 0)
                    {
                        buf += "Not-Present ";
                    }
                    buf += "), Sensortype " + util::str(stype);
                }
                break;
            case 2:
                {
                    // sensordata
                    uint16_t sdata = (data >> 8) | (data & 0xff) << 8;
                    buf += "Sensorvalue: Index " + util::str(sdata >> 12);
                    buf += ", Value " + util::str(sdata & 0xfff);
                }
                break;
            case 3:
                {
                    // hinkanalspiegel
                    buf += "Downstream Data " + util::hex(data, 4);
                }
                break;
            default:
                {
                    buf += "Unknown Subqualifier " + util::str(subqualifier);
                }
                break;
        }
    }
    else if (mainqualifier == 1 || mainqualifier == 0)
    {
        buf += "Device: " + util::str(device_id);
        buf += " (dSID:" + devid + ")";
        buf += ", Parameter Response: Offset " +
               util::hex(mainqualifier * 0x80 + subqualifier * 2, 2);
        buf += ", Value: " + util::hex(data, 4);
    }
    else
    {
        buf = "Unknown packet type from " + devid +
              "(address " + util::str(device_id) + ", data " +
              util::str(data) + ")";
    }

    return buf;
}

} // namespace
